"""Builder classes for defining functions and modules in HUGR."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from hugr import ops
from hugr.build.dfg import DefinitionBuilder, Function
from hugr.hugr import Hugr

if TYPE_CHECKING:
    from hugr.hugr.node_port import Node
    from hugr.metadata import NodeMetadata
    from hugr.tys import PolyFuncType, Type, TypeBound, TypeParam, TypeRow, Visibility

__all__ = ["Function", "Module"]


@dataclass
class Module(DefinitionBuilder[ops.Module]):
    """Build a top-level HUGR module.

    Examples:
        >>> m = Module()
        >>> m.hugr.entrypoint_op()
        Module()
    """

    hugr: Hugr[ops.Module]

    def __init__(self, hugr: Hugr | None = None) -> None:
        self.hugr = Hugr(ops.Module()) if hugr is None else hugr

    def define_main(self, input_types: TypeRow) -> Function:
        """Define the 'main' function in the module. See :meth:`define_function`."""
        return self.define_function("main", input_types)

    def define_function(
        self,
        name: str,
        input_types: TypeRow,
        output_types: TypeRow | None = None,
        type_params: list[TypeParam] | None = None,
        visibility: Visibility = "Private",
    ) -> Function:
        """Start building a function definition in the graph.

        Args:
            name: The name of the function.
            input_types: The input types for the function.
            output_types: The output types for the function.
                If not provided, it will be inferred after the function is built.
            type_params: The type parameters for the function, if polymorphic.
            visibility: the visibility of the function; defaults to "Private"

        Returns:
            The new function builder.
        """
        parent_op = ops.FuncDefn(
            name, input_types, type_params or [], visibility=visibility
        )
        func = Function.new_nested(parent_op, self.hugr, self.hugr.module_root)
        if output_types is not None:
            func.declare_outputs(output_types)
        return func

    def declare_function(
        self, name: str, signature: PolyFuncType, visibility: Visibility = "Public"
    ) -> Node:
        """Add a function declaration to the module.

        Args:
            name: The name of the function.
            signature: The (polymorphic) signature of the function.
            visibility: The visibility of the declaration; defaults to "Public"

        Returns:
            The node representing the function declaration.

        Examples:
            >>> from hugr.function import Module
            >>> m = Module()
            >>> sig = tys.PolyFuncType([], tys.FunctionType.empty())
            >>> m.declare_function("f", sig)
            Node(1)
        """
        return self.hugr.add_node(
            ops.FuncDecl(name, signature, visibility=visibility),
            self.hugr.entrypoint,
            num_outs=1,
        )

    def add_alias_defn(self, name: str, ty: Type) -> Node:
        """Add a type alias definition."""
        return self.hugr.add_node(ops.AliasDefn(name, ty), self.hugr.module_root)

    def add_alias_decl(self, name: str, bound: TypeBound) -> Node:
        """Add a type alias declaration."""
        return self.hugr.add_node(ops.AliasDecl(name, bound), self.hugr.module_root)

    @property
    def metadata(self) -> NodeMetadata:
        """Metadata associated with this module."""
        return self.hugr.entrypoint.metadata



from hugr import ops, tys, val
from hugr.build import Cfg, Dfg
from hugr.std.int import INT_T, DivMod, IntVal

from .conftest import validate


def build_basic_cfg(cfg: Cfg) -> None:
    with cfg.add_entry() as entry:
        entry.set_single_succ_outputs(*entry.inputs())
    cfg.branch_exit(entry[0])


def test_basic_cfg() -> None:
    cfg = Cfg(tys.Bool)
    build_basic_cfg(cfg)
    validate(cfg.hugr)


def test_branch() -> None:
    cfg = Cfg(tys.Bool, INT_T)
    entry = cfg.add_entry()
    entry.set_block_outputs(*entry.inputs())

    middle_1 = cfg.add_successor(entry[0])
    middle_1.set_single_succ_outputs(*middle_1.inputs())
    middle_2 = cfg.add_successor(entry[1])
    (i,) = middle_2.inputs()
    n = middle_2.add(DivMod(i, i))
    middle_2.set_single_succ_outputs(n[0])

    cfg.branch_exit(middle_1[0])
    cfg.branch_exit(middle_2[0])

    validate(cfg.hugr)


def test_nested_cfg() -> None:
    dfg = Dfg(tys.Bool)

    cfg = dfg.add_cfg(*dfg.inputs())

    build_basic_cfg(cfg)
    dfg.set_outputs(cfg)

    validate(dfg.hugr)


def test_dom_edge() -> None:
    cfg = Cfg(tys.Bool, tys.Unit, INT_T)
    with cfg.add_entry() as entry:
        b, u, i = entry.inputs()
        entry.set_block_outputs(b, i)

    # entry dominates both middles so Unit type can be used as inter-graph
    # value between basic blocks
    with cfg.add_successor(entry[0]) as middle_1:
        middle_1.set_block_outputs(u, *middle_1.inputs())

    with cfg.add_successor(entry[1]) as middle_2:
        middle_2.set_block_outputs(u, *middle_2.inputs())

    cfg.branch_exit(middle_1[0])
    cfg.branch_exit(middle_2[0])

    validate(cfg.hugr)


def test_asymm_types() -> None:
    # test different types going to entry block's successors
    with Cfg() as cfg:
        with cfg.add_entry() as entry:
            int_load = entry.load(IntVal(34))

            sum_ty = tys.Sum([[INT_T], [tys.Bool]])
            tagged_int = entry.add(ops.Tag(0, sum_ty)(int_load))
            entry.set_block_outputs(tagged_int)

        with cfg.add_successor(entry[0]) as middle:
            # discard the int and return the bool from entry
            middle.set_single_succ_outputs(middle.load(val.TRUE))

        # middle expects an int and exit expects a bool
        cfg.branch_exit(entry[1])
        cfg.branch_exit(middle[0])

    validate(cfg.hugr)


# TODO loop



from __future__ import annotations

import base64
import inspect
import sys
from abc import ABC, abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, Annotated, Any, Literal

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
    ValidationError,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    WrapValidator,
)
from pydantic_core import PydanticCustomError

from hugr.utils import deser_it

if TYPE_CHECKING:
    from collections.abc import Mapping


def _json_custom_error_validator(
    value: Any, handler: ValidatorFunctionWrapHandler, _info: ValidationInfo
) -> Any:
    """Simplify the error message to avoid a gross error stemming
    from exhaustive checking of all union options.

    As suggested at
    https://docs.pydantic.dev/latest/concepts/types/#named-recursive-types


    Used to define named recursive alias types.
    """
    return handler(value)
    try:
        return handler(value)
    except ValidationError as err:
        msg = "invalid_json"
        raise PydanticCustomError(
            msg,
            "Input is not valid json",
        ) from err


ExtensionId = str
ExtensionSet = list[ExtensionId]

default_model_config = ConfigDict()


class ConfiguredBaseModel(BaseModel):
    model_config = default_model_config

    @classmethod
    def update_model_config(cls, config: ConfigDict):
        cls.model_config.update(config)


# --------------------------------------------
# --------------- TypeParam ------------------
# --------------------------------------------


class BaseTypeParam(ABC, ConfiguredBaseModel):
    @abstractmethod
    def deserialize(self) -> tys.TypeParam: ...


class TypeTypeParam(BaseTypeParam):
    tp: Literal["Type"] = "Type"
    b: TypeBound

    def deserialize(self) -> tys.TypeTypeParam:
        return tys.TypeTypeParam(bound=self.b)


class BoundedNatParam(BaseTypeParam):
    tp: Literal["BoundedNat"] = "BoundedNat"
    bound: int | None

    def deserialize(self) -> tys.BoundedNatParam:
        return tys.BoundedNatParam(upper_bound=self.bound)


class StringParam(BaseTypeParam):
    tp: Literal["String"] = "String"

    def deserialize(self) -> tys.StringParam:
        return tys.StringParam()


class BytesParam(BaseTypeParam):
    tp: Literal["Bytes"] = "Bytes"

    def deserialize(self) -> tys.BytesParam:
        return tys.BytesParam()


class FloatParam(BaseTypeParam):
    tp: Literal["Float"] = "Float"

    def deserialize(self) -> tys.FloatParam:
        return tys.FloatParam()


class ListParam(BaseTypeParam):
    tp: Literal["List"] = "List"
    param: TypeParam

    def deserialize(self) -> tys.ListParam:
        return tys.ListParam(param=self.param.deserialize())


class TupleParam(BaseTypeParam):
    tp: Literal["Tuple"] = "Tuple"
    params: list[TypeParam]

    def deserialize(self) -> tys.TupleParam:
        return tys.TupleParam(params=deser_it(self.params))


class ConstParam(BaseTypeParam):
    tp: Literal["ConstType"] = "ConstType"
    ty: Type

    def deserialize(self) -> tys.ConstParam:
        return tys.ConstParam(ty=self.ty.deserialize())


class TypeParam(RootModel):
    """A type parameter."""

    root: Annotated[
        TypeTypeParam
        | BoundedNatParam
        | StringParam
        | FloatParam
        | BytesParam
        | ListParam
        | TupleParam
        | ConstParam,
        WrapValidator(_json_custom_error_validator),
    ] = Field(discriminator="tp")

    model_config = ConfigDict(json_schema_extra={"required": ["tp"]})

    def deserialize(self) -> tys.TypeParam:
        return self.root.deserialize()


# ------------------------------------------
# --------------- TypeArg ------------------
# ------------------------------------------


class BaseTypeArg(ABC, ConfiguredBaseModel):
    @abstractmethod
    def deserialize(self) -> tys.TypeArg: ...


class TypeTypeArg(BaseTypeArg):
    tya: Literal["Type"] = "Type"
    ty: Type

    def deserialize(self) -> tys.TypeTypeArg:
        return tys.TypeTypeArg(ty=self.ty.deserialize())


class BoundedNatArg(BaseTypeArg):
    tya: Literal["BoundedNat"] = "BoundedNat"
    n: int

    def deserialize(self) -> tys.BoundedNatArg:
        return tys.BoundedNatArg(n=self.n)


class StringArg(BaseTypeArg):
    tya: Literal["String"] = "String"
    arg: str

    def deserialize(self) -> tys.StringArg:
        return tys.StringArg(value=self.arg)


class FloatArg(BaseTypeArg):
    tya: Literal["Float"] = "Float"
    value: float

    def deserialize(self) -> tys.FloatArg:
        return tys.FloatArg(value=self.value)


class BytesArg(BaseTypeArg):
    tya: Literal["Bytes"] = "Bytes"
    value: str = Field(
        description="Base64-encoded byte string",
        json_schema_extra={"contentEncoding": "base64"},
    )

    def deserialize(self) -> tys.BytesArg:
        value = base64.b64decode(self.value)
        return tys.BytesArg(value=value)


class ListArg(BaseTypeArg):
    tya: Literal["List"] = "List"
    elems: list[TypeArg]

    def deserialize(self) -> tys.ListArg:
        return tys.ListArg(elems=deser_it(self.elems))


class ListConcatArg(BaseTypeArg):
    tya: Literal["ListConcat"] = "ListConcat"
    lists: list[TypeArg]

    def deserialize(self) -> tys.ListConcatArg:
        return tys.ListConcatArg(lists=deser_it(self.lists))


class TupleArg(BaseTypeArg):
    tya: Literal["Tuple"] = "Tuple"
    elems: list[TypeArg]

    def deserialize(self) -> tys.TupleArg:
        return tys.TupleArg(elems=deser_it(self.elems))


class TupleConcatArg(BaseTypeArg):
    tya: Literal["TupleConcat"] = "TupleConcat"
    tuples: list[TypeArg]

    def deserialize(self) -> tys.TupleConcatArg:
        return tys.TupleConcatArg(tuples=deser_it(self.tuples))


class VariableArg(BaseTypeArg):
    tya: Literal["Variable"] = "Variable"
    idx: int
    cached_decl: TypeParam

    def deserialize(self) -> tys.VariableArg:
        return tys.VariableArg(idx=self.idx, param=self.cached_decl.deserialize())


class TypeArg(RootModel):
    """A type argument."""

    root: Annotated[
        TypeTypeArg
        | BoundedNatArg
        | StringArg
        | BytesArg
        | FloatArg
        | ListArg
        | TupleArg
        | VariableArg,
        WrapValidator(_json_custom_error_validator),
    ] = Field(discriminator="tya")

    model_config = ConfigDict(json_schema_extra={"required": ["tya"]})

    def deserialize(self) -> tys.TypeArg:
        return self.root.deserialize()


# --------------------------------------------
# --------------- Container ------------------
# --------------------------------------------


class BaseType(ABC, ConfiguredBaseModel):
    @abstractmethod
    def deserialize(self) -> tys.Type: ...


class UnitSum(BaseType):
    """Simple sum type where all variants are empty tuples."""

    t: Literal["Sum"] = "Sum"
    s: Literal["Unit"] = "Unit"
    size: int

    def deserialize(self) -> tys.UnitSum:
        return tys.UnitSum(size=self.size)


class GeneralSum(BaseType):
    """General sum type that explicitly stores the types of the variants."""

    t: Literal["Sum"] = "Sum"
    s: Literal["General"] = "General"
    rows: list[TypeRow]

    def deserialize(self) -> tys.Sum:
        return tys.Sum(variant_rows=[[t.deserialize() for t in r] for r in self.rows])


class SumType(RootModel):
    root: Annotated[UnitSum | GeneralSum, Field(discriminator="s")]

    # This seems to be required for nested discriminated unions to work
    @property
    def t(self) -> str:
        return self.root.t

    model_config = ConfigDict(json_schema_extra={"required": ["s"]})

    def deserialize(self) -> tys.Sum | tys.UnitSum:
        return self.root.deserialize()


# ----------------------------------------------
# --------------- ClassicType ------------------
# ----------------------------------------------


class Variable(BaseType):
    """A type variable identified by an index into the array of TypeParams."""

    t: Literal["V"] = "V"
    i: int
    b: TypeBound

    def deserialize(self) -> tys.Variable:
        return tys.Variable(idx=self.i, bound=self.b)


class RowVar(BaseType):
    """A variable standing for a row of some (unknown) number of types.
    May occur only within a row; not a node input/output.
    """

    t: Literal["R"] = "R"
    i: int
    b: TypeBound

    def deserialize(self) -> tys.RowVariable:
        return tys.RowVariable(idx=self.i, bound=self.b)


class USize(BaseType):
    """Unsigned integer size type."""

    t: Literal["I"] = "I"

    def deserialize(self) -> tys.USize:
        return tys.USize()


class FunctionType(BaseType):
    """A graph encoded as a value. It contains a concrete signature and a set of
    required resources.
    """

    t: Literal["G"] = "G"

    input: TypeRow  # Value inputs of the function.
    output: TypeRow  # Value outputs of the function.

    @classmethod
    def empty(cls) -> FunctionType:
        return FunctionType(input=[], output=[])

    def deserialize(self) -> tys.FunctionType:
        return tys.FunctionType(
            input=deser_it(self.input),
            output=deser_it(self.output),
        )

    model_config = ConfigDict(
        # Needed to avoid random '\n's in the pydantic description
        json_schema_extra={
            "description": (
                "A graph encoded as a value. It contains a concrete signature and "
                "a set of required resources."
            )
        }
    )


class PolyFuncType(BaseType):
    """A polymorphic type scheme, i.e. of a FuncDecl, FuncDefn or OpDef.
    (Nodes/operations in the Hugr are not polymorphic.).
    """

    # The declared type parameters, i.e., these must be instantiated with the same
    # number of TypeArgs before the function can be called. This defines the indices
    # used for variables within the body.
    params: list[TypeParam]

    # Template for the function. May contain variables up to length of `params`
    body: FunctionType

    @classmethod
    def empty(cls) -> PolyFuncType:
        return PolyFuncType(params=[], body=FunctionType.empty())

    def deserialize(self) -> tys.PolyFuncType:
        return tys.PolyFuncType(
            params=deser_it(self.params),
            body=self.body.deserialize(),
        )

    model_config = ConfigDict(
        # Needed to avoid random '\n's in the pydantic description
        json_schema_extra={
            "description": (
                "A polymorphic type scheme, i.e. of a FuncDecl, FuncDefn or OpDef.  "
                "(Nodes/operations in the Hugr are not polymorphic.)"
            )
        }
    )


class TypeBound(Enum):
    Copyable = "C"
    Linear = "A"

    @staticmethod
    def join(*bs: TypeBound) -> TypeBound:
        """Computes the least upper bound for a sequence of bounds."""
        res = TypeBound.Copyable
        for b in bs:
            if b == TypeBound.Linear:
                return TypeBound.Linear
            if res == TypeBound.Copyable:
                res = b
        return res

    def __str__(self) -> str:
        match self:
            case TypeBound.Copyable:
                return "Copyable"
            case TypeBound.Linear:
                return "Linear"

    def __repr__(self) -> str:
        return f"TypeBound.{self}"


class Opaque(BaseType):
    """An opaque Type that can be downcasted by the extensions that define it."""

    t: Literal["Opaque"] = "Opaque"
    extension: ExtensionId
    id: str  # Unique identifier of the opaque type.
    args: list[TypeArg]
    bound: TypeBound

    def deserialize(self) -> tys.Opaque:
        return tys.Opaque(
            extension=self.extension,
            id=self.id,
            args=deser_it(self.args),
            bound=self.bound,
        )


class Alias(BaseType):
    """An Alias Type."""

    t: Literal["Alias"] = "Alias"
    bound: TypeBound
    name: str

    def deserialize(self) -> tys.Alias:
        return tys.Alias(name=self.name, bound=self.bound)


# ----------------------------------------------
# --------------- LinearType -------------------
# ----------------------------------------------


class Qubit(BaseType):
    """A qubit."""

    t: Literal["Q"] = "Q"

    def deserialize(self) -> tys._QubitDef:
        return tys.Qubit


class Type(RootModel):
    """A HUGR type."""

    root: Annotated[
        Qubit | Variable | RowVar | USize | FunctionType | SumType | Opaque | Alias,
        WrapValidator(_json_custom_error_validator),
        Field(discriminator="t"),
    ]

    model_config = ConfigDict(json_schema_extra={"required": ["t"]})

    def deserialize(self) -> tys.Type:
        return self.root.deserialize()


# -------------------------------------------
# --------------- TypeRow -------------------
# -------------------------------------------

TypeRow = list[Type]


# Now that all classes are defined, we need to update the ForwardRefs in all type
# annotations. We use some inspect magic to find all classes defined in this file
# and call model_rebuild()
classes = inspect.getmembers(
    sys.modules[__name__],
    lambda member: inspect.isclass(member) and member.__module__ == __name__,
)


def model_rebuild(
    classes: Mapping[str, type],
    config: ConfigDict | None = None,
    **kwargs,
):
    config = config or ConfigDict()
    for c in classes.values():
        if issubclass(c, ConfiguredBaseModel):
            c.update_model_config(config)
            c.model_rebuild(**kwargs)


model_rebuild(dict(classes))


from hugr import tys  # noqa: E402  # needed to avoid circular imports



from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING

import pytest

from hugr.hugr.base import Hugr
from hugr.passes._composable_pass import (
    ComposablePass,
    ComposedPass,
    PassResult,
    implement_pass_run,
)

if TYPE_CHECKING:
    from hugr.passes._scope import PassScope


def test_composable_pass() -> None:
    class DummyInlinePass(ComposablePass):
        def run(self, hugr: Hugr, inplace: bool = True) -> PassResult:
            return implement_pass_run(
                self,
                hugr=hugr,
                inplace=inplace,
                inplace_call=lambda hugr: PassResult.for_pass(
                    self,
                    hugr,
                    result=None,
                    inplace=True,
                    # Say that we modified the HUGR even though we didn't
                    modified=True,
                ),
            )

        def with_scope(self, scope: PassScope) -> DummyInlinePass:
            return self

    class DummyCopyPass(ComposablePass):
        def run(self, hugr: Hugr, inplace: bool = True) -> PassResult:
            return implement_pass_run(
                self,
                hugr=hugr,
                inplace=inplace,
                copy_call=lambda hugr: PassResult.for_pass(
                    self,
                    deepcopy(hugr),
                    result=None,
                    inplace=False,
                    # Say that we modified the HUGR even though we didn't
                    modified=True,
                ),
            )

        def with_scope(self, scope: PassScope) -> DummyCopyPass:
            return self

    dummy_inline = DummyInlinePass()
    dummy_copy = DummyCopyPass()

    composed_dummies = dummy_inline.then(dummy_copy)
    assert isinstance(composed_dummies, ComposedPass)

    assert dummy_inline.name == "DummyInlinePass"
    assert dummy_copy.name == "DummyCopyPass"
    assert composed_dummies.name == "Composed(DummyInlinePass, DummyCopyPass)"
    assert composed_dummies.then(dummy_inline).then(composed_dummies).name == (
        "Composed("
        + "DummyInlinePass, DummyCopyPass, "
        + "DummyInlinePass, "
        + "DummyInlinePass, DummyCopyPass)"
    )

    # Apply the passes
    hugr: Hugr = Hugr()
    new_hugr = composed_dummies(hugr, inplace=False)
    assert hugr == new_hugr
    assert new_hugr is not hugr

    # Verify the pass results
    hugr = Hugr()
    inplace_result = composed_dummies.run(hugr, inplace=True)
    assert inplace_result.modified
    assert inplace_result.inplace
    assert inplace_result.results == [
        ("DummyInlinePass", None),
        ("DummyCopyPass", None),
    ]
    assert inplace_result.hugr is hugr

    hugr = Hugr()
    copy_result = composed_dummies.run(hugr, inplace=False)
    assert copy_result.modified
    assert not copy_result.inplace
    assert copy_result.results == [
        ("DummyInlinePass", None),
        ("DummyCopyPass", None),
    ]
    assert copy_result.hugr is not hugr


def test_invalid_composable_pass() -> None:
    class DummyInvalidPass(ComposablePass):
        def run(self, hugr: Hugr, inplace: bool = True) -> PassResult:
            return implement_pass_run(
                self,
                hugr=hugr,
                inplace=inplace,
            )

        def with_scope(self, scope: PassScope) -> DummyInvalidPass:
            return self

    dummy_invalid = DummyInvalidPass()
    with pytest.raises(
        ValueError,
        match="DummyInvalidPass needs to implement at least an inplace or copy run method",  # noqa: E501
    ):
        dummy_invalid.run(Hugr())



from pathlib import Path

import pytest
import semver

from hugr import ops, tys
from hugr.build.dfg import Dfg
from hugr.build.function import Module
from hugr.envelope import EnvelopeConfig, EnvelopeFormat
from hugr.hugr.node_port import Node
from hugr.package import Package

from .conftest import QUANTUM_EXT, H


@pytest.fixture
def package() -> Package:
    mod = Module()
    f_id = mod.define_function("id", [tys.Qubit])
    f_id.set_outputs(f_id.input_node[0])

    mod2 = Module()
    f_id_decl = mod2.declare_function(
        "id", tys.PolyFuncType([], tys.FunctionType([tys.Qubit], [tys.Qubit]))
    )
    f_main = mod2.define_main([tys.Qubit])
    q = f_main.input_node[0]
    call = f_main.call(f_id_decl, q)
    f_main.set_outputs(call)
    return Package([mod.hugr, mod2.hugr])


@pytest.mark.parametrize(
    "compression", [None, 0], ids=["compression:None", "compression:0"]
)
@pytest.mark.parametrize(
    "format",
    [
        EnvelopeFormat.JSON,
        EnvelopeFormat.MODEL,
        EnvelopeFormat.MODEL_WITH_EXTS,
    ],
)
def test_envelope_binary(
    package: Package, compression: int | None, format: EnvelopeFormat
):
    # Binary compression roundtrip
    encoded = package.to_bytes(EnvelopeConfig(format=format, zstd=compression))
    decoded = Package.from_bytes(encoded)
    assert decoded == package


def test_envelope_text(package: Package):
    # String roundtrip
    encoded_str = package.to_str(EnvelopeConfig.TEXT)
    decoded = Package.from_str(encoded_str)
    assert decoded == package


def test_model(package: Package):
    model_pkg = package.to_model()

    # This value is statically defined in the rust bindings.
    assert model_pkg.version >= semver.Version(major=1, minor=0, patch=0)


def test_legacy_funcdefn():
    p = Path(__file__).parents[2] / "resources" / "test" / "hugr-no-visibility.hugr"
    try:
        with p.open("rb") as f:
            pkg_bytes = f.read()
    except FileNotFoundError:
        pytest.skip("Missing test file")
    decoded = Package.from_bytes(pkg_bytes)
    h = decoded.modules[0]
    op1 = h[Node(1)].op
    assert isinstance(op1, ops.FuncDecl)
    assert op1.visibility == "Public"
    op2 = h[Node(2)].op
    assert isinstance(op2, ops.FuncDefn)
    assert op2.visibility == "Private"


def test_model_import_with_ext():
    dfg = Dfg(tys.Qubit)
    h_outs = dfg.add_op(H, dfg.inputs()[0])
    dfg.set_outputs(h_outs)
    pkg = Package(modules=[dfg.hugr], extensions=[QUANTUM_EXT])
    data = pkg.to_bytes(config=EnvelopeConfig.BINARY)
    pkg1 = Package.from_bytes(data)
    data1 = pkg1.to_bytes(config=EnvelopeConfig.BINARY)
    pkg2 = Package.from_bytes(data1)
    assert pkg2.modules[0].num_nodes() == 8



from __future__ import annotations

import pytest

from hugr import tys
from hugr.build import dfg
from hugr.val import (
    FALSE,
    TRUE,
    Left,
    None_,
    Right,
    Some,
    Sum,
    Tuple,
    UnitSum,
    bool_value,
)

from .conftest import validate


def test_sums():
    assert Sum(0, tys.Tuple(), []) == Tuple()
    assert Sum(0, tys.Tuple(tys.Bool, tys.Bool), [TRUE, FALSE]) == Tuple(TRUE, FALSE)

    ty = tys.Sum([[], [tys.Bool, tys.Bool]])
    assert Sum(1, ty, [TRUE, FALSE]) == Some(TRUE, FALSE)
    assert Sum(1, ty, [TRUE, FALSE]) == Right([], [TRUE, FALSE])
    assert Sum(0, ty, []) == None_(tys.Bool, tys.Bool)
    assert Sum(0, ty, []) == Left([], [tys.Bool, tys.Bool])

    ty = tys.Sum([[tys.Bool], [tys.Bool]])
    assert Sum(0, ty, [TRUE]) == Left([TRUE], [tys.Bool])
    assert Sum(1, ty, [FALSE]) == Right([tys.Bool], [FALSE])

    assert Tuple() == Sum(0, tys.Tuple(), [])
    assert UnitSum(0, size=1) == Tuple()
    assert UnitSum(2, size=4) == Sum(2, tys.UnitSum(size=4), [])


@pytest.mark.parametrize(
    ("value", "string", "repr_str"),
    [
        (
            Sum(0, tys.Sum([[tys.Bool], [tys.Qubit], [tys.Bool]]), [TRUE]),
            "Sum(0, Sum([[Bool], [Qubit], [Bool]]), [TRUE])",
            "Sum(tag=0, typ=Sum([[Bool], [Qubit], [Bool]]), vals=[TRUE])",
        ),
        (UnitSum(0, size=1), "Unit", "Unit"),
        (UnitSum(0, size=2), "FALSE", "FALSE"),
        (UnitSum(1, size=2), "TRUE", "TRUE"),
        (UnitSum(2, size=5), "UnitSum(2, 5)", "UnitSum(2, 5)"),
        (Tuple(TRUE, FALSE), "Tuple(TRUE, FALSE)", "Tuple(TRUE, FALSE)"),
        (Some(TRUE, FALSE), "Some(TRUE, FALSE)", "Some(TRUE, FALSE)"),
        (None_(tys.Bool, tys.Bool), "None", "None(Bool, Bool)"),
        (
            Left([TRUE, FALSE], [tys.Bool]),
            "Left(TRUE, FALSE)",
            "Left(vals=[TRUE, FALSE], right_typ=[Bool])",
        ),
        (
            Right([tys.Bool, tys.Bool], [FALSE]),
            "Right(FALSE)",
            "Right(left_typ=[Bool, Bool], vals=[FALSE])",
        ),
    ],
)
def test_val_sum_str(value: Sum, string: str, repr_str: str):
    assert str(value) == string
    assert repr(value) == repr_str

    # Make sure the corresponding `Sum` also renders the same
    sum_val = Sum(value.tag, value.typ, value.vals)
    assert str(sum_val) == string
    assert repr(sum_val) == repr_str


def test_val_static_array():
    from hugr.std.collections.static_array import StaticArrayVal

    h = dfg.Dfg()
    load = h.load(
        StaticArrayVal([bool_value(x) for x in [True, False]], tys.Bool, "arr")
    )
    h.set_outputs(load)
    validate(h.hugr)



from pathlib import Path

import toml  # type: ignore[import-untyped]

import hugr


def test_versions_are_in_sync():
    """Checks if the pyproject.toml and package.__init__.py __version__ are in sync."""

    path = Path(__file__).resolve().parents[1] / "pyproject.toml"
    with Path.open(path, "r") as f:
        pyproject = toml.loads(f.read())
    pyproject_version = pyproject["project"]["version"]

    package_init_version = hugr.__version__

    assert package_init_version == pyproject_version



"""Typed metadata for HUGR nodes."""

from __future__ import annotations

import copy
from collections.abc import Mapping
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Protocol,
    TypeVar,
    overload,
)

import hugr._hugr.metadata as rust_metadata
from hugr.envelope import ExtensionDesc, GeneratorDesc

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

Meta = TypeVar("Meta")

# Type alias for json values.
# See <https://github.com/python/typing/issues/182#issuecomment-1320974824>
JsonType = str | int | float | bool | None | Mapping[str, "JsonType"] | list["JsonType"]


class Metadata(Protocol[Meta]):
    """Metadata for a HUGR node.

    This is a protocol for metadata entries that defines a unique key to
    identify the entry, and the type of the value.

    Values in a hugr are encoded using json. When the value type is not a
    primitive type, `to_json` and `from_json` must be implemented to serialize
    and deserialize the value.

    Args:
        value: The value of the metadata.
    """

    KEY: ClassVar[str]

    @classmethod
    def to_json(cls, value: Meta) -> JsonType:
        """Serialize the metadata value to a json value."""
        # Recursive types cannot be checked by isinstance, so we can only use a
        # surface level check.
        assert isinstance(
            value, str | int | float | bool | None | Mapping | list
        ), f"Cannot serialize {cls.KEY} metadata value {value} to json."
        return value

    @classmethod
    def from_json(cls, value: JsonType) -> Meta:
        """Deserialize the metadata value from the stored json value."""
        # Default implementation for primitive types.
        #
        # `Meta` not runtime checkable, so we need to use type: ignore.
        return value  # type: ignore  # noqa: PGH003


@dataclass
class NodeMetadata:
    """Key-value record of metadata for a HUGR node."""

    _dict: dict[str, JsonType]

    def __init__(self, metadata: dict[str, JsonType] | None = None) -> None:
        if metadata is None:
            metadata = {}
        # Only a shallow copy, values may still be shared with the original dictionary.
        self._dict = copy.copy(metadata)

    @overload
    def get(self, key: type[Metadata[Meta]], default: Meta) -> Meta: ...
    @overload
    def get(self, key: type[Metadata[Meta]], default: None = None) -> Meta | None: ...
    @overload
    def get(self, key: str, default: JsonType | None = None) -> JsonType | None: ...
    def get(
        self, key: str | type[Metadata[Meta]], default: Any | None = None
    ) -> Meta | JsonType | None:
        if isinstance(key, str):
            return self._dict.get(key, default)
        elif key.KEY in self._dict:
            val = self._dict[key.KEY]
            return key.from_json(val)
        else:
            return default

    def items(self) -> Iterable[tuple[str, JsonType]]:
        return self._dict.items()

    def as_dict(self) -> dict[str, JsonType]:
        return self._dict

    @overload
    def __getitem__(self, key: str) -> JsonType: ...
    @overload
    def __getitem__(self, key: type[Metadata[Meta]]) -> Meta: ...
    def __getitem__(self, key: str | type[Metadata[Meta]]) -> JsonType | Meta:
        if isinstance(key, str):
            return self._dict[key]
        else:
            val = self._dict[key.KEY]
            return key.from_json(val)

    @overload
    def __setitem__(self, key: str, value: JsonType) -> None: ...
    @overload
    def __setitem__(self, key: type[Metadata[Meta]], value: Meta) -> None: ...
    def __setitem__(
        self, key: str | type[Metadata[Meta]], value: JsonType | Meta
    ) -> None:
        if isinstance(key, str):
            # Recursive types cannot be checked by isinstance, so we can only use a
            # surface level check.
            assert isinstance(
                value, str | int | float | bool | None | Mapping | list
            ), f"Cannot set {key} metadata key with non-json value {value}."
            self._dict[key] = value
        else:
            # `value` must be a `Meta` type. Since `Meta` is not runtime checkable,
            # we need to use type: ignore.
            json_value = key.to_json(value)  # type: ignore  # noqa: PGH003
            self._dict[key.KEY] = json_value

    def __iter__(self) -> Iterator[str]:
        return iter(self._dict)

    def __len__(self) -> int:
        return len(self._dict)

    def __contains__(self, key: str | type[Metadata[Meta]]) -> bool:
        if not isinstance(key, str):
            key = key.KEY
        return key in self._dict

    def __repr__(self) -> str:
        return f"NodeMetadata({self._dict})"

    def __copy__(self) -> NodeMetadata:
        return NodeMetadata(copy.copy(self._dict))

    def __deepcopy__(self, memo: dict[int, JsonType]) -> NodeMetadata:
        return NodeMetadata(copy.deepcopy(self._dict, memo))


# --- Core metadata keys ---


class HugrGenerator(Metadata[GeneratorDesc]):
    """Metadata describing the generator that defined the HUGR module.

    This value is only valid when set at the module root node.
    """

    KEY = rust_metadata.HUGR_GENERATOR

    @classmethod
    def to_json(cls, value: GeneratorDesc) -> dict[str, str]:
        return value.to_json()

    @classmethod
    def from_json(cls, value: JsonType) -> GeneratorDesc:
        return GeneratorDesc.from_json(value)


class HugrUsedExtensions(Metadata[list[ExtensionDesc]]):
    """Metadata storing the list of extensions required to define the HUGR.

    This list may contain additional extensions that are no longer present in
    the Hugr.

    This value is only valid when set at the module root node.
    """

    KEY = rust_metadata.HUGR_USED_EXTENSIONS

    @classmethod
    def to_json(cls, value: list[ExtensionDesc]) -> JsonType:
        return [e.to_json() for e in value]

    @classmethod
    def from_json(cls, value: JsonType) -> list[ExtensionDesc]:
        if not isinstance(value, list):
            msg = (
                "Expected UsedExtensions metadata to be a list,"
                + f" but got {type(value)}"
            )
            raise TypeError(msg)
        return [ExtensionDesc.from_json(e) for e in value]



"""Floating point types and operations."""

from __future__ import annotations

from dataclasses import dataclass

import hugr.model as model
from hugr import ext, val
from hugr.std._util import _load_extension

FLOAT_TYPES_EXTENSION = _load_extension("arithmetic.float.types")

FLOAT_T = FLOAT_TYPES_EXTENSION.types["float64"].instantiate([])


FLOAT_OPS_EXTENSION = _load_extension("arithmetic.float")


@dataclass
class FloatVal(val.ExtensionValue):
    """Custom value for a floating point number."""

    v: float

    def to_value(self) -> val.Extension:
        name = "ConstF64"
        payload = {"value": self.v}
        return val.Extension(name, typ=FLOAT_T, val=payload)

    def __str__(self) -> str:
        return f"{self.v}"

    def _resolve_used_extensions_inplace(
        self, registry: ext.ExtensionRegistry | None = None
    ) -> ext.ExtensionResolutionResult:
        result = ext.ExtensionResolutionResult()
        result.used_extensions.add_extension(FLOAT_TYPES_EXTENSION)
        return result

    def to_model(self) -> model.Term:
        return model.Apply("arithmetic.float.const_f64", [model.Literal(self.v)])



"""A Protocol for a composable pass.

Currently unstable.
"""

from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Protocol, runtime_checkable

if TYPE_CHECKING:
    from collections.abc import Callable

    from hugr.hugr.base import Hugr
    from hugr.passes._scope import PassScope


# Type alias for a pass name
PassName = str


@runtime_checkable
class ComposablePass(Protocol):
    """A Protocol which represents a composable Hugr transformation."""

    def __call__(self, hugr: Hugr, *, inplace: bool = True) -> Hugr:
        """Call the pass to transform a HUGR, returning a Hugr."""
        return self.run(hugr, inplace=inplace).hugr

    def run(self, hugr: Hugr, *, inplace: bool = True) -> PassResult:
        """Run the pass to transform a HUGR, returning a PassResult.

        See :func:`implement_pass_run` for a helper function to implement this method.
        """
        ...

    def with_scope(self, scope: PassScope) -> ComposablePass:
        """Set the scope configuration for the pass.

        As of `hugr 0.14.*`, this configuration is only a guidance, and may be
        ignored by the pass.

        In the future, passes will be required to respect the scope configuration.
        """
        # For hugr passes, this is tracked by <https://github.com/Quantinuum/hugr/issues/2771>
        ...

    @property
    def name(self) -> PassName:
        """Returns the name of the pass."""
        return self.__class__.__name__

    def then(self, other: ComposablePass) -> ComposablePass:
        """Perform another composable pass after this pass."""
        return ComposedPass(self, other)


def implement_pass_run(
    composable_pass: ComposablePass,
    *,
    hugr: Hugr,
    inplace: bool,
    inplace_call: Callable[[Hugr], PassResult] | None = None,
    copy_call: Callable[[Hugr], PassResult] | None = None,
) -> PassResult:
    """Helper function to implement a ComposablePass.run method, given an
    inplace or copy-returning pass method.

    At least one of the `inplace_call` or `copy_call` arguments must be provided.

    :param composable_pass: The pass being run. Used for error messages.
    :param hugr: The Hugr to apply the pass to.
    :param inplace: Whether to apply the pass inplace.
    :param inplace_call: The method to apply the pass inplace.
    :param copy_call: The method to apply the pass by copying the Hugr.
    :return: The result of the pass application.
    :raises ValueError: If neither `inplace_call` nor `copy_call` is provided.
    """
    if inplace:
        if inplace_call is not None:
            return inplace_call(hugr)
        elif copy_call is not None:
            pass_result = copy_call(hugr)
            if pass_result.modified:
                hugr._overwrite_hugr(pass_result.hugr)
                pass_result.inplace = True
            pass_result.hugr = hugr
            return pass_result
    elif not inplace:
        if copy_call is not None:
            return copy_call(hugr)
        elif inplace_call is not None:
            new_hugr = deepcopy(hugr)
            pass_result = inplace_call(new_hugr)
            pass_result.inplace = False
            return pass_result

    msg = (
        f"{composable_pass.name} needs to implement at least "
        + "an inplace or copy run method"
    )
    raise ValueError(msg)


@dataclass
class ComposedPass(ComposablePass):
    """A sequence of composable passes."""

    passes: list[ComposablePass]

    def __init__(self, *passes: ComposablePass) -> None:
        self.passes = []
        for composable_pass in passes:
            if isinstance(composable_pass, ComposedPass):
                self.passes.extend(composable_pass.passes)
            else:
                self.passes.append(composable_pass)

    def run(self, hugr: Hugr, *, inplace: bool = True) -> PassResult:
        def apply(inplace: bool, hugr: Hugr) -> PassResult:
            pass_result = PassResult(hugr=hugr, inplace=inplace)
            for comp_pass in self.passes:
                new_result = comp_pass.run(pass_result.hugr, inplace=inplace)
                pass_result = pass_result.then(new_result)
            return pass_result

        return implement_pass_run(
            self,
            hugr=hugr,
            inplace=inplace,
            inplace_call=lambda hugr: apply(True, hugr),
            copy_call=lambda hugr: apply(False, hugr),
        )

    def with_scope(self, scope: PassScope) -> ComposedPass:
        """Set the scope configuration for the composed pass."""
        return ComposedPass(
            *[composable_pass.with_scope(scope) for composable_pass in self.passes]
        )

    @property
    def name(self) -> PassName:
        names = [composable_pass.name for composable_pass in self.passes]
        return f"Composed({ ', '.join(names) })"


@dataclass
class PassResult:
    """The result of a series of composed passes applied to a HUGR.

    Includes a flag indicating whether the passes modified the HUGR, and an
    arbitrary result object for each pass.

    :attr hugr: The transformed Hugr.
    :attr inplace: Whether the pass was applied inplace.
        If this is `True`, `hugr` will be the same object passed as input.
        If this is `False`, `hugr` will be an independent copy of the original Hugr.
    :attr modified: Whether the pass made changes to the HUGR.
        If `False`, `hugr` will have the same contents as the original Hugr.
        If `True`, no guarantees are made about the contents of `hugr`.
    :attr results: The result of each applied pass, as a tuple of the pass name
        and the result.
    """

    hugr: Hugr
    inplace: bool = False
    modified: bool = False
    results: list[tuple[PassName, Any]] = field(default_factory=list)

    @classmethod
    def for_pass(
        cls,
        composable_pass: ComposablePass,
        hugr: Hugr,
        *,
        result: Any,
        inplace: bool,
        modified: bool = True,
    ) -> PassResult:
        """Create a new PassResult after a pass application.

        :param hugr: The Hugr that was transformed.
        :param composable_pass: The pass that was applied.
        :param result: The result of the pass application.
        :param inplace: Whether the pass was applied inplace.
        :param modified: Whether the pass modified the HUGR.
        """
        return cls(
            hugr=hugr,
            inplace=inplace,
            modified=modified,
            results=[(composable_pass.name, result)],
        )

    def then(self, other: PassResult) -> PassResult:
        """Extend the PassResult with the results of another PassResult.

        Keeps the hugr returned by the last pass.
        """
        return PassResult(
            hugr=other.hugr,
            inplace=self.inplace and other.inplace,
            modified=self.modified or other.modified,
            results=self.results + other.results,
        )



from collections.abc import Iterable
from dataclasses import dataclass

import pytest

from hugr import tys, val
from hugr.build import Module
from hugr.hugr import Hugr, Node
from hugr.passes._scope import GlobalScope, InScope, LocalScope, PassScope


@dataclass(frozen=True)
class TestHugr:
    hugr: Hugr
    module_root: Node
    public_func: Node
    public_func_nested: Node
    private_func: Node
    public_func_decl: Node
    private_func_decl: Node
    const_def: Node


@pytest.fixture
def test_hugr() -> TestHugr:
    mod = Module()
    module_root = mod.hugr.module_root

    public_func = mod.define_function("public_func", [], [], visibility="Public")
    with public_func.add_nested() as public_func_nested:
        pass  # set outputs?

    private_func = mod.define_function("private_func", [], [])

    public_func_decl = mod.declare_function(
        "public_func_decl", tys.PolyFuncType([], tys.FunctionType([], []))
    )

    private_func_decl = mod.declare_function(
        "private_func_decl",
        tys.PolyFuncType([], tys.FunctionType([tys.Qubit], [tys.Qubit])),
        visibility="Private",
    )

    const_def = mod.add_const(val.TRUE)

    return TestHugr(
        hugr=mod.hugr,
        module_root=module_root,
        public_func=public_func.parent_node,
        public_func_nested=public_func_nested.parent_node,
        private_func=private_func.parent_node,
        public_func_decl=public_func_decl,
        private_func_decl=private_func_decl,
        const_def=const_def,
    )


@pytest.mark.parametrize(
    ("scope", "recursive"), [(LocalScope.FLAT, False), (LocalScope.RECURSIVE, True)]
)
def test_local_scope(test_hugr: TestHugr, scope: PassScope, recursive: bool):
    assert scope.recursive() == recursive

    # When the entrypoint is the module root,
    # the pass should not be applied to any regions.
    test_hugr.hugr.entrypoint = test_hugr.module_root
    assert scope.root(test_hugr.hugr) is None
    assert list(scope.regions(test_hugr.hugr)) == []
    for n, _ in test_hugr.hugr.nodes():
        assert scope.in_scope(test_hugr.hugr, n) == InScope.NO

    # Public function with a nested DFG
    test_hugr.hugr.entrypoint = test_hugr.public_func
    assert scope.root(test_hugr.hugr) == test_hugr.public_func
    expected_regions = (
        [test_hugr.public_func, test_hugr.public_func_nested]
        if recursive
        else [test_hugr.public_func]
    )
    assert list(scope.regions(test_hugr.hugr)) == expected_regions

    assert scope.in_scope(test_hugr.hugr, test_hugr.module_root) == InScope.NO
    assert (
        scope.in_scope(test_hugr.hugr, test_hugr.public_func)
        == InScope.PRESERVE_INTERFACE
    )

    assert scope.in_scope(test_hugr.hugr, test_hugr.public_func_nested) == InScope.YES

    for n in [
        test_hugr.module_root,
        test_hugr.private_func,
        test_hugr.public_func_decl,
        test_hugr.private_func_decl,
        test_hugr.const_def,
    ]:
        assert scope.in_scope(test_hugr.hugr, n) == InScope.NO

    # DFG inside a function
    test_hugr.hugr.entrypoint = test_hugr.public_func_nested
    assert scope.root(test_hugr.hugr) == test_hugr.public_func_nested
    assert list(scope.regions(test_hugr.hugr)) == [test_hugr.public_func_nested]

    for n in [
        test_hugr.module_root,
        test_hugr.public_func,
        test_hugr.private_func,
        test_hugr.public_func_decl,
        test_hugr.private_func_decl,
        test_hugr.const_def,
    ]:
        assert scope.in_scope(test_hugr.hugr, n) == InScope.NO
    assert (
        scope.in_scope(test_hugr.hugr, test_hugr.public_func_nested)
        == InScope.PRESERVE_INTERFACE
    )


def test_preserve_all(test_hugr: TestHugr):
    preserve = [
        test_hugr.public_func,
        test_hugr.private_func,
        test_hugr.public_func_decl,
        test_hugr.private_func_decl,
        test_hugr.const_def,
    ]
    check_preserve(test_hugr, GlobalScope.PRESERVE_ALL, preserve)


def check_preserve(
    test_hugr: TestHugr, scope: GlobalScope, expected_chs: Iterable[Node]
):
    assert scope.recursive()
    expected_chs = set(expected_chs)
    assert scope.root(test_hugr.hugr) == test_hugr.module_root
    assert set(scope.regions(test_hugr.hugr)) == {
        test_hugr.public_func,
        test_hugr.private_func,
        test_hugr.public_func_nested,
    }

    assert (
        scope.in_scope(test_hugr.hugr, test_hugr.module_root)
        == InScope.PRESERVE_INTERFACE
    )

    for n in [
        test_hugr.public_func,
        test_hugr.private_func,
        test_hugr.public_func_decl,
        test_hugr.public_func_nested,
        test_hugr.private_func_decl,
        test_hugr.const_def,
    ]:
        expected = InScope.PRESERVE_INTERFACE if n in expected_chs else InScope.YES
        assert scope.in_scope(test_hugr.hugr, n) == expected, f"{n} among {test_hugr}"

    expected_chs.add(test_hugr.module_root)  # Things to preserve
    assert set(scope.preserve_interface(test_hugr.hugr)) == expected_chs


def test_preserve_public(test_hugr: TestHugr):
    preserve = [test_hugr.public_func, test_hugr.public_func_decl]
    check_preserve(test_hugr, GlobalScope.PRESERVE_PUBLIC, preserve)


def test_preserve_entrypoint(test_hugr: TestHugr):
    test_hugr.hugr.entrypoint = test_hugr.hugr.module_root
    preserve = [test_hugr.public_func, test_hugr.public_func_decl]
    check_preserve(test_hugr, GlobalScope.PRESERVE_ENTRYPOINT, preserve)

    test_hugr.hugr.entrypoint = test_hugr.public_func_nested
    preserve = [test_hugr.public_func_nested]
    check_preserve(test_hugr, GlobalScope.PRESERVE_ENTRYPOINT, preserve)



# ruff: noqa: I001

import hugr
from hugr.package import Package
import hugr.ops as ops
import hugr.tys as tys
from hugr import ext, val

from hugr.build import Dfg
from hugr.std.collections.list import List
from hugr.std.float import FLOAT_TYPES_EXTENSION
from hugr.std.int import CONVERSIONS_EXTENSION, INT_T, INT_TYPES_EXTENSION
import pytest

from .conftest import H, QUANTUM_EXT, TEST_EXT, TEST_TYPE_OPAQUE, TEST_OP_OPAQUE


def test_extension_ops() -> None:
    h = Dfg(tys.Qubit)
    (q,) = h.inputs()

    h_node = h.add(H(q))
    h.set_outputs(h_node.out(0))

    op_exts = h.hugr[h_node].op.used_extensions().ids()
    assert set(op_exts) == {QUANTUM_EXT.name, FLOAT_TYPES_EXTENSION.name, "prelude"}

    exts = h.hugr.used_extensions().ids()
    assert set(exts) == {QUANTUM_EXT.name, FLOAT_TYPES_EXTENSION.name, "prelude"}


def test_core_ops() -> None:
    h = Dfg(INT_T, tys.Bool)
    a, b = h.inputs()

    tuple_n = h.add(ops.MakeTuple()(a, b))
    h.set_outputs(tuple_n.out(0))

    op_exts = h.hugr[tuple_n].op.used_extensions().ids()
    assert set(op_exts) == {INT_TYPES_EXTENSION.name, "prelude"}

    exts = h.hugr.used_extensions().ids()
    assert set(exts) == {INT_TYPES_EXTENSION.name, "prelude"}


def test_transitive_extension_dependencies() -> None:
    op = ops.ExtOp(CONVERSIONS_EXTENSION.get_op("itousize"), args=[])
    exts = op.used_extensions().ids()
    assert CONVERSIONS_EXTENSION.name in exts
    assert FLOAT_TYPES_EXTENSION.name in exts


def test_func_calls() -> None:
    h = Dfg(tys.Bool)
    [a] = h.inputs()

    f_decl = h.module_root_builder().declare_function(
        "f",
        tys.PolyFuncType(
            params=[tys.TypeTypeParam(bound=tys.TypeBound.Copyable)],
            body=tys.FunctionType([tys.Qubit], []),
        ),
    )

    call = h.call(
        f_decl,
        a,
        instantiation=tys.FunctionType([tys.Qubit], []),
        type_args=[tys.TypeTypeArg(INT_T)],
    )
    h.set_outputs()

    op_exts = h.hugr[call].op.used_extensions().ids()
    assert set(op_exts) == {INT_TYPES_EXTENSION.name, "prelude"}

    exts = h.hugr.used_extensions().ids()
    assert set(exts) == {INT_TYPES_EXTENSION.name, "prelude"}


def test_type_arguments() -> None:
    list_ty = List(tys.Qubit)
    h = Dfg(list_ty)
    (items,) = h.inputs()
    h.set_outputs(items)

    exts = h.hugr.used_extensions().ids()
    assert set(exts) == {hugr.std.collections.list.EXTENSION.name, "prelude"}


def test_op_signature_contains_same_extension() -> None:
    """Test that used_extensions() works when an op's signature contains
    types from the same extension as the operation itself.

    Regression test for https://github.com/Quantinuum/hugr/issues/2829
    """

    # Create an extension with a custom type
    my_ext = ext.Extension("test.self_referential", ext.Version(0, 1, 0))
    my_type_def = my_ext.add_type_def(
        ext.TypeDef(
            name="MyType",
            description="A custom type",
            params=[],
            bound=ext.ExplicitBound(tys.TypeBound.Copyable),
        )
    )

    # Create an ExtType from this extension's type definition
    my_type = tys.ExtType(my_type_def, args=[])

    # Add an operation to the same extension that uses MyType in its signature
    my_ext.add_op_def(
        ext.OpDef(
            name="ProcessMyType",
            description="An operation that uses MyType",
            signature=ext.OpDefSig(tys.FunctionType([my_type], [my_type])),
        )
    )

    # Create an ExtOp from this definition
    op = ops.ExtOp(my_ext.get_op("ProcessMyType"), args=[])

    # This would have failed with ExtensionRegistry.ExtensionExists before the fix
    # because the signature contains my_type (from my_ext) and we also add my_ext
    exts = op.used_extensions()
    assert my_ext.name in exts.ids()


@pytest.mark.parametrize(
    ("typ", "extensions"),
    [
        # Opaque type
        (TEST_TYPE_OPAQUE, [TEST_EXT]),
        (tys.Sum([[TEST_TYPE_OPAQUE], [tys.Bool]]), [TEST_EXT]),
        (tys.Tuple(TEST_TYPE_OPAQUE, tys.Bool), [TEST_EXT]),
        (tys.Option(TEST_TYPE_OPAQUE), [TEST_EXT]),
        (tys.Either([TEST_TYPE_OPAQUE], [tys.Bool]), [TEST_EXT]),
        (tys.UnitSum(3), []),
        (tys.FunctionType([TEST_TYPE_OPAQUE], [tys.Bool]), [TEST_EXT]),
        (
            tys.PolyFuncType(
                [tys.FloatParam()], tys.FunctionType([TEST_TYPE_OPAQUE], [tys.Bool])
            ),
            [TEST_EXT],
        ),
        (tys.Variable(0, tys.TypeBound.Copyable), []),
        (tys.RowVariable(0, tys.TypeBound.Copyable), []),
        (tys.Alias("MyAlias", tys.TypeBound.Copyable), []),
    ],
)
def test_type_resolution(typ: tys.Type, extensions: list[ext.Extension]) -> None:
    """Test that Opaque types are tracked and resolved correctly."""
    _, result = typ._resolve_used_extensions()
    for extension in extensions:
        assert extension.name in result.unresolved_extensions
    if extensions:
        assert (TEST_EXT.name, "TestType") in result.unresolved_types
    assert not result.unresolved_ops

    test_registry = ext.ExtensionRegistry()
    for extension in extensions:
        test_registry.add_extension(extension)
    _, result = typ._resolve_used_extensions(test_registry)
    for extension in extensions:
        assert extension.name in result.used_extensions.ids()
    assert not result.unresolved_types
    assert not result.unresolved_ops


@pytest.mark.parametrize(
    ("arg", "extensions"),
    [
        (tys.TypeTypeArg(TEST_TYPE_OPAQUE), [TEST_EXT]),
        (tys.ListArg([tys.TypeTypeArg(TEST_TYPE_OPAQUE)]), [TEST_EXT]),
        (
            tys.TupleArg([tys.TypeTypeArg(TEST_TYPE_OPAQUE), tys.BoundedNatArg(5)]),
            [TEST_EXT],
        ),
        (
            tys.ListConcatArg([tys.ListArg([tys.TypeTypeArg(TEST_TYPE_OPAQUE)])]),
            [TEST_EXT],
        ),
        (
            tys.TupleConcatArg([tys.TupleArg([tys.TypeTypeArg(TEST_TYPE_OPAQUE)])]),
            [TEST_EXT],
        ),
        (tys.VariableArg(0, tys.ConstParam(TEST_TYPE_OPAQUE)), [TEST_EXT]),
        (tys.BoundedNatArg(42), []),
        (tys.StringArg("hello"), []),
        (tys.FloatArg(3.14), []),
        (tys.BytesArg(b"bytes"), []),
    ],
)
def test_type_arg_resolution(arg: tys.TypeArg, extensions: list[ext.Extension]) -> None:
    """Test that type arguments are tracked and resolved correctly."""
    _, result = arg._resolve_used_extensions()
    for extension in extensions:
        assert extension.name in result.unresolved_extensions
    if extensions:
        assert (TEST_EXT.name, "TestType") in result.unresolved_types
    assert not result.unresolved_ops

    test_registry = ext.ExtensionRegistry()
    for extension in extensions:
        test_registry.add_extension(extension)
    _, result = arg._resolve_used_extensions(test_registry)
    for extension in extensions:
        assert extension.name in result.used_extensions.ids()
    assert not result.unresolved_types
    assert not result.unresolved_ops


@pytest.mark.parametrize(
    ("param", "extensions"),
    [
        (tys.ListParam(tys.ConstParam(TEST_TYPE_OPAQUE)), [TEST_EXT]),
        (
            tys.TupleParam([tys.ConstParam(TEST_TYPE_OPAQUE), tys.BoundedNatParam()]),
            [TEST_EXT],
        ),
        (tys.ConstParam(TEST_TYPE_OPAQUE), [TEST_EXT]),
        (tys.TypeTypeParam(tys.TypeBound.Copyable), []),
        (tys.BoundedNatParam(100), []),
        (tys.StringParam(), []),
        (tys.FloatParam(), []),
        (tys.BytesParam(), []),
    ],
)
def test_type_param_resolution(
    param: tys.TypeParam, extensions: list[ext.Extension]
) -> None:
    """Test that type parameters are tracked and resolved correctly."""
    _, result = param._resolve_used_extensions()
    for extension in extensions:
        assert extension.name in result.unresolved_extensions
    if extensions:
        assert (TEST_EXT.name, "TestType") in result.unresolved_types
    assert not result.unresolved_ops

    test_registry = ext.ExtensionRegistry()
    for extension in extensions:
        test_registry.add_extension(extension)
    _, result = param._resolve_used_extensions(test_registry)
    for extension in extensions:
        assert extension.name in result.used_extensions.ids()
    assert not result.unresolved_types
    assert not result.unresolved_ops


@pytest.mark.parametrize(
    ("op", "extensions"),
    [
        (TEST_OP_OPAQUE, [TEST_EXT]),
        (ops.MakeTuple([TEST_TYPE_OPAQUE]), [TEST_EXT]),
        (ops.UnpackTuple([TEST_TYPE_OPAQUE]), [TEST_EXT]),
        (ops.Tag(0, tys.Sum([[TEST_TYPE_OPAQUE], []])), [TEST_EXT]),
        (ops.LoadConst(TEST_TYPE_OPAQUE), [TEST_EXT]),
        (ops.Conditional(tys.Sum([[TEST_TYPE_OPAQUE], []]), [tys.Bool]), [TEST_EXT]),
        (ops.Case([TEST_TYPE_OPAQUE]), [TEST_EXT]),
        (ops.TailLoop([TEST_TYPE_OPAQUE], [tys.Bool]), [TEST_EXT]),
        (ops.DFG([TEST_TYPE_OPAQUE]), [TEST_EXT]),
        (ops.CFG([TEST_TYPE_OPAQUE]), [TEST_EXT]),
        (ops.DataflowBlock([TEST_TYPE_OPAQUE]), [TEST_EXT]),
        (ops.ExitBlock([TEST_TYPE_OPAQUE]), [TEST_EXT]),
        (ops.FuncDefn("test_fn", [TEST_TYPE_OPAQUE]), [TEST_EXT]),
        (
            ops.FuncDecl(
                "test_fn",
                tys.PolyFuncType([], tys.FunctionType([TEST_TYPE_OPAQUE], [])),
            ),
            [TEST_EXT],
        ),
        (ops.CallIndirect(tys.FunctionType([TEST_TYPE_OPAQUE], [])), [TEST_EXT]),
        (ops.AliasDefn("MyAlias", TEST_TYPE_OPAQUE), [TEST_EXT]),
    ],
)
def test_op_resolution(op: ops.Op, extensions: list[ext.Extension]) -> None:
    """Test that ops are tracked and resolved correctly."""
    _, result = op._resolve_used_extensions()
    for extension in extensions:
        assert extension.name in result.unresolved_extensions

    if isinstance(op, ops.Custom):
        assert (op.extension, op.op_name) in result.unresolved_ops
        assert not result.unresolved_types
    elif extensions:
        assert not result.unresolved_ops
        assert (TEST_EXT.name, "TestType") in result.unresolved_types
    else:
        assert not result.unresolved_ops
        assert not result.unresolved_types

    test_ext_registry = ext.ExtensionRegistry()
    for extension in extensions:
        test_ext_registry.add_extension(extension)
    _, result = op._resolve_used_extensions(test_ext_registry)
    for extension in extensions:
        assert extension.name in result.used_extensions.ids()
    assert not result.unresolved_ops
    assert not result.unresolved_types


@pytest.mark.parametrize(
    ("value", "extensions"),
    [
        (val.Sum(0, tys.Sum([[TEST_TYPE_OPAQUE], []]), []), [TEST_EXT]),
        (val.Tuple(val.TRUE), []),
        (val.Some(val.TRUE), []),
        (val.None_(TEST_TYPE_OPAQUE), [TEST_EXT]),
        (val.Left([], [TEST_TYPE_OPAQUE]), [TEST_EXT]),
        (val.Right([TEST_TYPE_OPAQUE], []), [TEST_EXT]),
        (val.UnitSum(0, 3), []),
        (val.Extension("test", TEST_TYPE_OPAQUE, {}), [TEST_EXT]),
    ],
)
def test_value_resolution(value: val.Value, extensions: list[ext.Extension]) -> None:
    """Test that values are tracked and resolved correctly."""
    result = value._resolve_used_extensions_inplace()
    for extension in extensions:
        assert extension.name in result.unresolved_extensions

    # Reset the value for the second test by creating a fresh copy
    test_ext_registry = ext.ExtensionRegistry()
    for extension in extensions:
        test_ext_registry.add_extension(extension)
    result = value._resolve_used_extensions_inplace(test_ext_registry)
    for extension in extensions:
        assert extension.name in result.used_extensions.ids()


def test_hugr_with_opaque_type_resolve() -> None:
    """Test that Hugr.used_extensions can resolve opaque types."""
    # Create a DFG with the opaque type
    h = Dfg(TEST_TYPE_OPAQUE)
    (inp,) = h.inputs()
    h.set_outputs(inp)

    # Without resolve_from, unresolved extensions are tracked
    result_no_resolve = h.hugr.used_extensions()
    assert TEST_EXT.name in result_no_resolve.unresolved_extensions

    # With resolve_from, should work
    test_ext_registry = ext.ExtensionRegistry()
    test_ext_registry.add_extension(TEST_EXT)
    exts = h.hugr.used_extensions(resolve_from=test_ext_registry)
    assert TEST_EXT.name in exts.used_extensions.ids()


def test_lower_funcs_resolve() -> None:
    """Test that extensions in `OpDef.lower_func`s are resolved correctly."""
    # A custom extension used by inside lower function.
    inner_op_def = ext.OpDef(
        name="InnerOp",
        description="inner custom op",
        signature=ext.OpDefSig(tys.FunctionType.endo([tys.Bool])),
    )
    inner_ext = ext.Extension(
        version=ext.Version(0, 1, 0),
        name="inner",
        types={},
    )
    inner_ext.add_op_def(inner_op_def)

    # Build a simple identity HUGR to use as a lowering implementation.
    dfg = Dfg(tys.Bool)
    [b] = dfg.inputs()
    b = dfg.add_op(inner_ext.get_op("InnerOp").instantiate(args=[]), b).out(0)
    dfg.set_outputs(b)
    lower_hugr = dfg.hugr

    # Wrap it in a FixedHugr lower function.
    fixed_hugr = ext.FixedHugr(extensions=[inner_ext.name], hugr=lower_hugr)

    # Build an extension whose op carries the lower_func.
    outer_op_def = ext.OpDef(
        name="OuterOp",
        description="outer op with lowering",
        signature=ext.OpDefSig(tys.FunctionType.endo([tys.Bool])),
        lower_funcs=[fixed_hugr],
    )
    outer_ext = ext.Extension(
        version=ext.Version(0, 1, 0),
        name="outer",
        types={},
    )
    outer_ext.add_op_def(outer_op_def)

    # Include the extension in a package and check that the lower func is resolved.
    package = Package(
        modules=[],
        extensions=[outer_ext],
    )
    used_exts = package.used_extensions()
    assert outer_ext.name in used_exts.used_extensions
    assert inner_ext.name in used_exts.used_extensions

    bytes = package.to_bytes()
    package = Package.from_bytes(bytes, extensions=used_exts.used_extensions)
    assert len(package.extensions[0].operations["OuterOp"].lower_funcs) == 1

    used_exts = package.used_extensions()
    assert outer_ext.name in used_exts.used_extensions.ids()
    assert inner_ext.name in used_exts.used_extensions.ids()



import re
from collections import Counter

import pytest

# test deprecated module
from hugr.qsystem.result import REG_INDEX_PATTERN, QsysResult, QsysShot


@pytest.mark.parametrize(
    ("identifier", "match"),
    [
        ("sadfj", None),
        ("asdf_sdf", None),
        ("asdf3h32", None),
        ("dsf[3]asdf", None),
        ("_s34fd_fd[12]", None),
        ("afsd3[34]sdf", None),
        ("asdf[2]", ("asdf", 2)),
        ("as3df[21234]", ("as3df", 21234)),
        ("as3ABdfAB[2]", ("as3ABdfAB", 2)),
    ],
)
def test_reg_index_pattern_match(identifier, match: tuple[str, int] | None):
    """Test regex pattern matches tags indexing in to registers."""
    mtch = re.match(REG_INDEX_PATTERN, identifier)
    if mtch is None:
        assert match is None
        return
    parsed = (mtch.group(1), int(mtch.group(2)))
    assert parsed == match


def test_as_dict():
    results = QsysShot()
    results.append("tag1", 1)
    results.append("tag2", 2)
    results.append("tag2", 3)
    assert results.as_dict() == {"tag1": 1, "tag2": 3}


def test_to_register_bits():
    results = QsysShot()
    results.append("c[0]", 1)
    results.append("c[1]", 0)
    results.append("c[3]", 1)
    results.append("d", [1, 0, 1, 0])
    results.append("x[5]", 1)
    results.append("x", 0)

    assert results.to_register_bits() == {"c": "1001", "d": "1010", "x": "0"}

    shots = QsysResult([results, results])
    assert shots.register_counts() == {
        "c": Counter({"1001": 2}),
        "d": Counter({"1010": 2}),
        "x": Counter({"0": 2}),
    }


@pytest.mark.parametrize(
    "results",
    [
        QsysShot([("t", 1.0)]),
        QsysShot([("t[1]", 1.0)]),
        QsysShot([("t", [1.0])]),
        QsysShot([("t[0]", [0])]),
        QsysShot([("t[0]", 3)]),
    ],
)
def test_to_register_bits_bad(results: QsysShot):
    with pytest.raises(ValueError, match="Expected bit"):
        _ = results.to_register_bits()


def test_counter():
    shot1 = QsysShot()
    shot1.append("c", [1, 0, 1, 0])
    shot1.append("d", [1, 0, 1])

    shot2 = QsysShot()
    shot2.append("c", [1, 0, 1])

    shots = QsysResult([shot1, shot2])
    assert shots.register_counts() == {
        "c": Counter({"1010": 1, "101": 1}),
        "d": Counter({"101": 1}),
    }
    with pytest.raises(ValueError, match="same length"):
        _ = shots.register_counts(strict_lengths=True)

    with pytest.raises(ValueError, match="All shots must have the same registers"):
        _ = shots.register_counts(strict_names=True)


def test_pytket():
    """Test that results observing strict tagging conventions can be converted to pytket
    shot results."""
    pytest.importorskip("pytket", reason="pytket not installed")

    hsim_shots = QsysResult(
        ([("c", [1, 0]), ("d", [1, 0, 0])], [("c", [0, 0]), ("d", [1, 0, 1])])
    )

    pytket_result = hsim_shots.to_pytket()
    from pytket._tket.unit_id import Bit
    from pytket.backends.backendresult import BackendResult
    from pytket.utils.outcomearray import OutcomeArray

    bits = [Bit("c", 0), Bit("c", 1), Bit("d", 0), Bit("d", 1), Bit("d", 2)]
    expected = BackendResult(
        c_bits=bits,
        shots=OutcomeArray.from_readouts([[1, 0, 1, 0, 0], [0, 0, 1, 0, 1]]),
    )

    assert pytket_result == expected


def test_collate_tag():
    # test use of same tag for all entries of array

    shotlist = []
    for _ in range(10):
        shot = QsysShot()
        _ = [
            shot.append(reg, 1)
            for reg, size in (("c", 3), ("d", 5))
            for _ in range(size)
        ]
        shotlist.append(shot)

    weird_shot = QsysShot((("c", 1), ("d", 1), ("d", 0), ("e", 1)))
    assert weird_shot.collate_tags() == {"c": [1], "d": [1, 0], "e": [1]}

    lst_shot = QsysShot([("lst", [1, 0, 1]), ("lst", [1, 0, 1])])
    shots = QsysResult([*shotlist, weird_shot, lst_shot])

    counter = shots.collated_counts()
    assert counter == Counter(
        {
            (("c", "111"), ("d", "11111")): 10,
            (("c", "1"), ("d", "10"), ("e", "1")): 1,
            (("lst", "101101"),): 1,
        }
    )

    float_shots = QsysResult(
        [QsysShot([("f", 1.0), ("f", 0.1)]), QsysShot([("f", [2.0]), ("g", 2.0)])]
    )

    assert float_shots.collated_shots() == [
        {"f": [1.0, 0.1]},
        {"f": [[2.0]], "g": [2.0]},
    ]


def test_collated_digitstring_counts():
    shotlist = []
    for _ in range(10):
        shot = QsysShot()
        _ = [
            shot.append(reg, 3)
            for reg, size in (("c", 3), ("d", 5))
            for _ in range(size)
        ]
        shotlist.append(shot)

    int_shot = QsysShot((("c", 1), ("d", 2), ("d", 0), ("e", 4)))
    assert int_shot.collate_tags() == {"c": [1], "d": [2, 0], "e": [4]}

    lst_shot = QsysShot([("lst", [1, 0, 2]), ("lst", [1, 0, 2])])
    shots = QsysResult([*shotlist, int_shot, lst_shot])

    counter = shots.collated_digitstring_counts()
    assert counter == Counter(
        {
            (("c", "333"), ("d", "33333")): 10,
            (("c", "1"), ("d", "20"), ("e", "4")): 1,
            (("lst", "102102"),): 1,
        }
    )


@pytest.mark.parametrize(
    "value",
    [2, -1, 1.0, [1, 2, 0]],
)
def test_collated_counts_bad_values(value):
    """Test that collated_counts raises ValueError for non-bit values."""
    shots = QsysResult([QsysShot([("a", value)])])
    with pytest.raises(ValueError, match="Expected bit"):
        _ = shots.collated_counts()


@pytest.mark.parametrize(
    "value",
    [10, -1, 1.0, [1, 15, 0]],
)
def test_collated_digitstring_counts_bad_values(value):
    """Test that collated_digitstring_counts raises ValueError for non-digit values."""
    shots = QsysResult([QsysShot([("a", value)])])
    with pytest.raises(ValueError, match="Expected single digit"):
        _ = shots.collated_digitstring_counts()


def test_qsys_shot_sequence_behavior():
    """Test that QsysShot implements Sequence protocol correctly."""
    shot = QsysShot([("a", 1), ("b", 2), ("c", 3)])

    # Test __len__
    assert len(shot) == 3

    # Test __getitem__ with int index
    assert shot[0] == ("a", 1)
    assert shot[1] == ("b", 2)
    assert shot[-1] == ("c", 3)

    # Test __getitem__ with slice
    assert shot[0:2] == [("a", 1), ("b", 2)]
    assert shot[1:] == [("b", 2), ("c", 3)]

    # Test __iter__
    entries = list(shot)
    assert entries == [("a", 1), ("b", 2), ("c", 3)]

    # Test iteration with for loop
    result = []
    for tag, value in shot:
        result.append((tag, value))
    assert result == [("a", 1), ("b", 2), ("c", 3)]


def test_qsys_result_sequence_behavior():
    """Test that QsysResult implements Sequence protocol correctly."""
    shot1 = QsysShot([("a", 1), ("b", 2)])
    shot2 = QsysShot([("c", 3), ("d", 4)])
    shot3 = QsysShot([("e", 5)])

    result = QsysResult([shot1, shot2, shot3])

    # Test __len__
    assert len(result) == 3

    # Test __getitem__ with int index
    assert result[0] == shot1
    assert result[1] == shot2
    assert result[-1] == shot3

    # Test __getitem__ with slice
    assert result[0:2] == [shot1, shot2]
    assert result[1:] == [shot2, shot3]

    # Test __iter__
    shots = list(result)
    assert shots == [shot1, shot2, shot3]

    # Test iteration behavior
    assert all(isinstance(shot, QsysShot) for shot in result)



"""Visualise HUGR using graphviz."""

import html
from collections.abc import Iterable
from dataclasses import dataclass, field

import graphviz as gv  # type: ignore[import-untyped]
from graphviz import Digraph
from typing_extensions import assert_never

from hugr.hugr import Hugr
from hugr.ops import AsExtOp, Case
from hugr.tys import CFKind, ConstKind, FunctionKind, Kind, OrderKind, ValueKind

from .node_port import InPort, Node, OutPort


@dataclass(frozen=True)
class Palette:
    """A set of colours used for rendering."""

    background: str
    node: str
    edge: str
    entrypoint_edge: str
    dark: str
    const: str
    discard: str
    node_border: str
    port_border: str

    @classmethod
    def named(cls, name: str) -> "Palette":
        return PALETTE[name]


PALETTE: dict[str, Palette] = {
    "default": Palette(
        background="white",
        node="#ACCBF9",
        edge="#1CADE4",
        entrypoint_edge="#F4A261",
        dark="black",
        const="#77CEEF",
        discard="#ff8888",
        node_border="white",
        port_border="#1CADE4",
    ),
    "nb": Palette(
        background="white",
        node="#7952B3",
        edge="#FFC107",
        entrypoint_edge="#00CFC1",
        dark="#343A40",
        const="#7c55b4",
        discard="#ff8888",
        node_border="#9d80c7",
        port_border="#ffd966",
    ),
    "zx": Palette(
        background="white",
        node="#629DD1",
        edge="#297FD5",
        entrypoint_edge="#FF8243",
        dark="#112D4E",
        const="#a1eea1",
        discard="#ff8888",
        node_border="#D8F8D8",
        port_border="#E8A5A5",
    ),
}


@dataclass
class RenderConfig:
    """Configuration for rendering a HUGR to a graphviz dot file."""

    #: The palette to use for rendering. See :obj:`PALETTE` for the included options.
    palette: Palette = field(default_factory=lambda: PALETTE["default"])
    #: If true prepend extension name to operation name.
    qualify_op_name: bool = False
    #: If true display node metadata.
    display_metadata: bool = True
    #: If true display node numbering.
    display_node_id: bool = False
    #: If true display full type name on the edges.
    display_link_label: bool = True
    #: Max length for edge labels. None means no truncation.
    max_edge_label_length: int | None = 24
    #: Max length for node labels. None means no truncation.
    max_node_label_length: int | None = 24
    #: Max length for metadata display. None means no truncation.
    max_metadata_length: int | None = 20


class DotRenderer:
    """Render a HUGR to a graphviz dot file.

    Args:
        config: Render config
    """

    config: RenderConfig

    def __init__(self, config: RenderConfig | None = None) -> None:
        self.config = config or RenderConfig()
        self.nodes: set[Node] = set()

    def render(self, hugr: Hugr, root: Node | None = None) -> Digraph:
        """Render a HUGR to a graphviz dot object.

        Args:
            hugr: The HUGR to render.
            root: Root node defining the set of nodes to render. By default this is the
                module root and all nodes are rendered. If this is a container node, all
                nodes under it are rendered. Every incoming edge to the rendered set and
                outgoing edge from it is also shown, with its other endpoint labelled
                with its node index.
        """
        root = root or hugr.module_root
        graph_attr = {
            "rankdir": "",
            "ranksep": "0.1",
            "nodesep": "0.15",
            "margin": "0",
            "bgcolor": self.config.palette.background,
        }
        if name := hugr[root].metadata.get("name", None):
            name = html.escape(str(name))
        else:
            name = ""

        graph = gv.Digraph(name, strict=False)
        graph.attr(**graph_attr)

        self._viz_node(root, hugr, graph)

        for src_port, tgt_port in hugr.links():
            kind = hugr.port_kind(src_port)
            self._viz_link(src_port, tgt_port, kind, graph)

        return graph

    def store(
        self, hugr: Hugr, filename: str, format: str = "svg", root: Node | None = None
    ) -> None:
        """Render a HUGR and save it to a file.

        Args:
            hugr: The HUGR to render.
            filename: Filename for saving the rendered graph.
            format: The format used for rendering ('pdf', 'png', etc.).
                Defaults to SVG.
            root: Root node defining the set of nodes to render. By default this is the
                module root and all nodes are rendered. If this is a container node, all
                nodes under it are rendered. Every incoming edge to the rendered set and
                outgoing edge from it is also shown, with its other endpoint labelled
                with its node index.
        """
        gv_graph = self.render(hugr, root=root)
        gv_graph.render(filename, format=format)

    _FONTFACE = "monospace"

    _HTML_LABEL_TEMPLATE = """
    <TABLE BORDER="{border_width}" CELLBORDER="0" CELLSPACING="1" CELLPADDING="1"
        BGCOLOR="{node_back_color}" COLOR="{border_colour}">
    {inputs_row}
    <TR>
        <TD>
        <TABLE BORDER="0" CELLBORDER="0">
            <TR><TD><FONT POINT-SIZE="{fontsize}" FACE="{fontface}"
                COLOR="{label_color}"><B>{node_label}</B>{node_data}</FONT></TD></TR>
        </TABLE>
        </TD>
    </TR>
    {outputs_row}
    </TABLE>
    """

    _HTML_PORTS_ROW_TEMPLATE = """
        <TR>
            <TD>
                <TABLE BORDER="0" CELLBORDER="0" CELLSPACING="3" CELLPADDING="2">
                    <TR>
                        {port_cells}
                    </TR>
                </TABLE>
            </TD>
        </TR>
    """

    _HTML_PORT_TEMPLATE = (
        '<TD BGCOLOR="{back_colour}" COLOR="{border_colour}"'
        ' PORT="{port_id}" BORDER="{border_width}">'
        '<FONT POINT-SIZE="10.0" FACE="{fontface}" COLOR="{font_colour}">{port}</FONT>'
        "</TD>"
    )

    _INPUT_PREFIX = "in."
    _OUTPUT_PREFIX = "out."

    def _format_html_label(self, **kwargs: str) -> str:
        _HTML_LABEL_DEFAULTS = {
            "label_color": self.config.palette.dark,
            "node_back_color": self.config.palette.node,
            "inputs_row": "",
            "outputs_row": "",
            "border_colour": self.config.palette.port_border,
            "border_width": "1",
            "fontface": self._FONTFACE,
            "fontsize": 11.0,
        }
        return self._HTML_LABEL_TEMPLATE.format(**{**_HTML_LABEL_DEFAULTS, **kwargs})

    def _html_ports(self, ports: Iterable[str], id_prefix: str) -> str:
        return self._HTML_PORTS_ROW_TEMPLATE.format(
            port_cells="".join(
                self._HTML_PORT_TEMPLATE.format(
                    port=port,
                    # differentiate input and output node identifiers
                    # with a prefix
                    port_id=id_prefix + port,
                    back_colour=self.config.palette.background,
                    font_colour=self.config.palette.dark,
                    border_width="1",
                    border_colour=self.config.palette.port_border,
                    fontface=self._FONTFACE,
                )
                for port in ports
            )
        )

    def _in_port_name(self, p: InPort) -> str:
        return f"{p.node.idx}:{self._INPUT_PREFIX}{p.offset}"

    def _out_port_name(self, p: OutPort) -> str:
        return f"{p.node.idx}:{self._OUTPUT_PREFIX}{p.offset}"

    def _in_order_name(self, n: Node) -> str:
        return f"{n.idx}:{self._INPUT_PREFIX}None"

    def _out_order_name(self, n: Node) -> str:
        return f"{n.idx}:{self._OUTPUT_PREFIX}None"

    def _viz_node(
        self,
        node: Node,
        hugr: Hugr,
        graph: Digraph,
        *,
        sibling_order: int | None = None,
    ) -> None:
        """Render a (possibly nested) node to a graphviz graph.

        Args:
            node: The node to render.
            hugr: The HUGR to render.
            graph: The graphviz graph to render the node to.
            sibling_order: The order of the node in the region's sibling list.
        """
        meta = hugr[node].metadata
        if len(meta) > 0 and self.config.display_metadata:
            if self.config.max_metadata_length is not None:
                data = "<BR/><BR/>" + "<BR/>".join(
                    html.escape(key)
                    + ": "
                    + html.escape(
                        _smart_truncate(str(value), self.config.max_metadata_length)
                    )
                    for key, value in meta.items()
                )
            else:
                data = "<BR/><BR/>" + "<BR/>".join(
                    html.escape(key) + ": " + html.escape(str(value))
                    for key, value in meta.items()
                )
        else:
            data = ""

        in_ports = [str(i) for i in range(hugr.num_in_ports(node))]
        out_ports = [str(i) for i in range(hugr.num_out_ports(node))]
        inputs_row = (
            self._html_ports(in_ports, self._INPUT_PREFIX) if len(in_ports) > 0 else ""
        )
        outputs_row = (
            self._html_ports(out_ports, self._OUTPUT_PREFIX)
            if len(out_ports) > 0
            else ""
        )

        op = hugr[node].op
        if isinstance(op, AsExtOp) and not self.config.qualify_op_name:
            op_name = op.op_def().name
        elif isinstance(op, Case) and sibling_order is not None:
            # Indicate the case number
            op_name = f"{op.name()}[{sibling_order}]"
        else:
            op_name = op.name()

        if self.config.max_node_label_length is not None:
            op_name = _smart_truncate(
                op_name, max_length=self.config.max_node_label_length
            )

        op_name = html.escape(op_name)
        node_label = (
            f"{op_name} <BR/>({node.idx}) " if self.config.display_node_id else op_name
        )

        label_config = {
            "node_back_color": self.config.palette.node,
            "node_label": node_label,
            "node_data": data,
            "inputs_row": inputs_row,
            "outputs_row": outputs_row,
            "border_colour": self.config.palette.background,
            "border_width": "1",
        }
        if hugr.children(node):
            # Some overrides when rendering a container node
            label_config["node_back_color"] = self.config.palette.edge
            label_config["border_colour"] = self.config.palette.port_border
        if node == hugr.entrypoint:
            label_config["node_label"] = "<b>[" + label_config["node_label"] + "]</b>"
            label_config["border_colour"] = self.config.palette.entrypoint_edge
            label_config["border_width"] = "2"

        if hugr.children(node):
            with graph.subgraph(name=f"cluster{node.idx}") as sub:
                for sibling_order, child in enumerate(hugr.children(node)):
                    self._viz_node(child, hugr, sub, sibling_order=sibling_order)
                html_label = self._format_html_label(**label_config)
                sub.node(f"{node.idx}", shape="plain", label=f"<{html_label}>")
                sub.attr(
                    label="",
                    margin="10",
                    color=label_config["border_colour"],
                    penwidth=label_config["border_width"],
                )
        else:
            html_label = self._format_html_label(**label_config)
            graph.node(f"{node.idx}", label=f"<{html_label}>", shape="plain")
        self.nodes.add(node)

    def _viz_link(
        self, src_port: OutPort, tgt_port: InPort, kind: Kind, graph: Digraph
    ) -> None:
        edge_attr = {
            "penwidth": "1.5",
            "arrowhead": "none",
            "arrowsize": "1.0",
            "fontname": self._FONTFACE,
            "fontsize": "9",
            "fontcolor": "black",
        }

        label = ""
        match kind:
            case ValueKind(ty):
                if self.config.display_link_label:
                    # TODO: Figure out why some nested types result in all type fields
                    # being printed instead of just the `__str__` representation.
                    label = html.escape(str(ty))
                color = self.config.palette.edge
            case OrderKind():
                color = self.config.palette.dark
            case ConstKind() | FunctionKind():
                color = self.config.palette.const
            case CFKind():
                color = self.config.palette.dark
            case _:
                assert_never(kind)

        src = self._out_port_name(src_port)
        tgt = self._in_port_name(tgt_port)

        unknown_src = src_port.node not in self.nodes
        unknown_tgt = tgt_port.node not in self.nodes
        if unknown_src and unknown_tgt:
            return
        if unknown_src:
            src = f"{src_port.node.idx}"
            html_label = self._format_html_label(
                node_back_color=self.config.palette.node,
                node_label=f"{src_port.node}",
                node_data="",
                border_colour=self.config.palette.background,
                border_width="1",
            )
            graph.node(src, label=f"<{html_label}>", shape="plain")
        if unknown_tgt:
            tgt = f"{tgt_port.node.idx}"
            html_label = self._format_html_label(
                node_back_color=self.config.palette.node,
                node_label=f"{tgt_port.node}",
                node_data="",
                border_colour=self.config.palette.background,
                border_width="1",
            )
            graph.node(tgt, label=f"<{html_label}>", shape="plain")

        if self.config.max_edge_label_length is not None:
            label = _smart_truncate(
                label=label, max_length=self.config.max_edge_label_length
            )

        graph.edge(src, tgt, xlabel=label, color=color, **edge_attr)


def _smart_truncate(label: str, max_length: int = 24) -> str:
    """Truncate a string for display.

    Rules:
    - If the string fits, return it unchanged.
        - If truncation is needed, truncate only at delimiters (space or bracket chars).
    - Never truncate the first word (the initial token before a delimiter), even if it
      exceeds `max_length`.
        - If truncation happens inside unmatched brackets, append enough closing
            brackets after the ellipsis to close any still-open ones.
            Supported bracket pairs are `()`, `{}`, `[]`, `<>`.
    """
    if len(label) <= max_length:
        return label

    ellipsis = "..."

    delimiters = {
        " ",
        "(",
        ")",
        "{",
        "}",
        "[",
        "]",
        "<",
        ">",
        ".",
    }
    opening_brackets = {"(", "{", "[", "<"}

    # Identify the protected prefix (never truncate this), which is at least the
    # first token, and optionally also the next token if the string starts with
    # `Token<token` / `Token(token` / etc.
    first_delim_idx = next(
        (i for i, ch in enumerate(label) if ch in delimiters),
        len(label),
    )
    protected_end = first_delim_idx
    if first_delim_idx < len(label) and label[first_delim_idx] in opening_brackets:
        second_start = first_delim_idx + 1
        second_delim_idx = next(
            (
                i
                for i, ch in enumerate(label[second_start:], start=second_start)
                if ch in delimiters
            ),
            len(label),
        )
        protected_end = second_delim_idx

    prefix = label[:protected_end].rstrip()

    if prefix == label:
        return label

    truncated = prefix
    bracket_pairs = [("(", ")"), ("{", "}"), ("[", "]"), ("<", ">")]
    # we first try to close parenthesis without the ellipsis
    # and check if it is the same as the original label
    for open_ch, close_ch in bracket_pairs:
        balance = prefix.count(open_ch) - prefix.count(close_ch)
        if balance > 0:
            truncated += close_ch * balance

    if truncated == label:
        return label
    # If not, we add the ellipsis and then close any still-open parenthesis
    truncated_w_ellipsis = prefix + ellipsis

    for open_ch, close_ch in bracket_pairs:
        balance = prefix.count(open_ch) - prefix.count(close_ch)
        if balance > 0:
            truncated_w_ellipsis += close_ch * balance
    return truncated_w_ellipsis



from __future__ import annotations

from typing import Any

from semver import Version

from hugr.envelope import EnvelopeConfig, ExtensionDesc, GeneratorDesc
from hugr.hugr import Hugr
from hugr.metadata import HugrGenerator, HugrUsedExtensions, JsonType, Metadata


class CustomMetadata(Metadata[list[JsonType]]):
    KEY = "custom.metadata"


def test_metadata_properties() -> None:
    hugr = Hugr[Any]()
    node = hugr[hugr.module_root]

    gen = GeneratorDesc(name="hugr-py-test", version=Version.parse("1.2.3"))
    node.metadata[HugrGenerator] = gen

    assert HugrGenerator in node.metadata
    assert HugrGenerator.KEY in node.metadata
    assert HugrUsedExtensions not in node.metadata
    assert HugrUsedExtensions.KEY not in node.metadata

    assert list(node.metadata) == [HugrGenerator.KEY]
    assert (
        str(node.metadata)
        == f"NodeMetadata({{'{HugrGenerator.KEY}': {gen.to_json()}}})"
    )


def test_metadata_roundtrip() -> None:
    hugr = Hugr[Any]()

    gen = GeneratorDesc(name="hugr-py-test", version=Version.parse("1.2.3"))
    exts = [
        ExtensionDesc(name="ext.a", version=Version.parse("0.1.0")),
        ExtensionDesc(name="ext.b", version=Version.parse("2.0.0")),
    ]

    # Set the metadata on the module root node
    node = hugr[hugr.module_root]
    node.metadata[HugrGenerator] = gen
    node.metadata[HugrUsedExtensions] = exts

    custom_payload: list[JsonType] = [1, "payload", True, None, {"key": "value"}]
    node.metadata[CustomMetadata.KEY] = custom_payload
    node.metadata[CustomMetadata] = custom_payload

    # Roundtrip serialization
    data = hugr.to_bytes(EnvelopeConfig.TEXT)
    loaded = Hugr.from_bytes(data)
    node = loaded[loaded.module_root]

    # Typed readback
    assert node.metadata[HugrGenerator] == gen
    assert node.metadata.get(HugrGenerator) == gen
    assert node.metadata[HugrUsedExtensions] == exts
    assert node.metadata.get(HugrUsedExtensions) == exts
    assert node.metadata[CustomMetadata] == custom_payload
    assert node.metadata[CustomMetadata.KEY] == custom_payload
    assert node.metadata.get(CustomMetadata.KEY) == custom_payload
    assert node.metadata.get(CustomMetadata.KEY, "default") == custom_payload

    # Check the raw JSON encoding
    raw = node.metadata.as_dict()
    assert raw[HugrGenerator.KEY] == {"name": "hugr-py-test", "version": "1.2.3"}
    assert raw[HugrUsedExtensions.KEY] == [
        {"name": "ext.a", "version": "0.1.0"},
        {"name": "ext.b", "version": "2.0.0"},
    ]
    assert raw["custom.metadata"] == custom_payload


def test_metadata_default() -> None:
    hugr = Hugr[Any]()
    node = hugr[hugr.module_root]

    assert node.metadata.get(HugrGenerator) is None
    assert node.metadata.get(
        HugrGenerator, GeneratorDesc("hugr-py-test", Version.parse("1.2.3"))
    ) == GeneratorDesc("hugr-py-test", Version.parse("1.2.3"))
    assert node.metadata.get("missing.metadata") is None
    assert node.metadata.get("missing.metadata", [1, 2, 3]) == [1, 2, 3]



"""Extension loading utilities."""

import pkgutil

from hugr._serialization.extension import Extension as PdExtension
from hugr.ext import Extension


def _load_extension(name: str) -> Extension:
    replacement = name.replace(".", "/")
    json_str = pkgutil.get_data(__name__, f"_json_defs/{replacement}.json")
    assert json_str is not None
    return PdExtension.model_validate_json(json_str).deserialize()



"""HUGR pointer operations."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from hugr import tys
from hugr.std._util import _load_extension

if TYPE_CHECKING:
    from hugr.ext import ExtensionRegistry, ExtensionResolutionResult

EXTENSION = _load_extension("ptr")

PTR_T_DEF = EXTENSION.types["ptr"]


@dataclass(eq=False)
class Ptr(tys.ExtType):
    """Pointer type with a fixed element type."""

    def __init__(self, ty: tys.Type) -> None:
        ty_arg = tys.TypeTypeArg(ty)

        self.type_def = EXTENSION.types["ptr"]
        self.args = [ty_arg]

    @property
    def ty(self) -> tys.Type:
        """Returns the type of the pointer."""
        assert isinstance(
            self.args[0], tys.TypeTypeArg
        ), "Pointer elements must have a valid type"
        return self.args[0].ty

    def _resolve_used_extensions(
        self, registry: ExtensionRegistry | None = None
    ) -> tuple[Ptr, ExtensionResolutionResult]:
        ext_type, result = super()._resolve_used_extensions(registry)

        assert isinstance(
            ext_type, tys.ExtType
        ), "HUGR internal error, expected resolved type to be extension type."
        assert (
            ext_type.type_def == EXTENSION.types["ptr"]
        ), "HUGR internal error, expected resolved type to be pointer."

        ptr = Ptr(tys.Unit)
        ptr.args = ext_type.args
        return ptr, result



"""`hugr` is a Python package for the Quantinuum HUGR common
representation.
"""

from . import model
from .hugr.base import Hugr
from .hugr.node_port import Direction, InPort, Node, OutPort, Wire

__all__ = [
    "Hugr",
    "Node",
    "OutPort",
    "InPort",
    "Direction",
    "Wire",
    "model",
]

# This is updated by our release-please workflow, triggered by this
# annotation: x-release-please-version
__version__ = "0.15.5"



import pytest

from hugr.hugr import Node, OutPort


def test_index():
    n = Node(0, _num_out_ports=3)
    assert n[0] == OutPort(n, 0)
    assert n[1] == OutPort(n, 1)
    assert n[2] == OutPort(n, 2)
    assert n[-1] == OutPort(n, 2)

    with pytest.raises(IndexError, match="Index 3 out of range"):
        _ = n[3]

    with pytest.raises(IndexError, match="Index -8 out of range"):
        _ = n[-8]


def test_slices():
    n = Node(0, _num_out_ports=3)
    all_ports = [OutPort(n, i) for i in range(3)]

    assert list(n) == all_ports
    assert list(n[:0]) == []
    assert list(n[0:0]) == []
    assert list(n[0:1]) == [OutPort(n, 0)]
    assert list(n[1:2]) == [OutPort(n, 1)]
    assert list(n[:]) == all_ports
    assert list(n[0:]) == all_ports
    assert list(n[:3]) == all_ports
    assert list(n[0:3]) == all_ports
    assert list(n[0:999]) == all_ports
    assert list(n[999:1000]) == []
    assert list(n[-1:]) == [OutPort(n, 2)]
    assert list(n[-3:]) == all_ports

    with pytest.raises(IndexError, match="Index -4 out of range"):
        _ = n[-4:]

    n0 = Node(0, _num_out_ports=0)
    assert list(n0) == []
    assert list(n0[:0]) == []
    assert list(n0[:10]) == []
    assert list(n0[0:0]) == []
    assert list(n0[0:]) == []
    assert list(n0[10:]) == []
    assert list(n0[:]) == []



"""Helpers to import hugr graphs from hugr model to their python representation."""

import json
from collections.abc import Generator, Iterable
from dataclasses import dataclass, field
from typing import cast

import hugr.model as model
from hugr import val
from hugr.hugr import InPort, OutPort
from hugr.hugr.base import Hugr
from hugr.hugr.node_port import Node
from hugr.metadata import NodeMetadata
from hugr.ops import (
    CFG,
    DFG,
    AliasDecl,
    AliasDefn,
    Call,
    CallIndirect,
    Case,
    Conditional,
    Const,
    Custom,
    DataflowBlock,
    ExitBlock,
    FuncDecl,
    FuncDefn,
    Input,
    LoadConst,
    LoadFunc,
    MakeTuple,
    Op,
    Output,
    Tag,
    TailLoop,
    UnpackTuple,
)
from hugr.std.collections.array import ArrayVal
from hugr.std.float import FloatVal
from hugr.std.int import IntVal
from hugr.tys import (
    BoundedNatArg,
    BoundedNatParam,
    BytesArg,
    BytesParam,
    ConstParam,
    FloatArg,
    FloatParam,
    FunctionType,
    ListArg,
    ListConcatArg,
    ListParam,
    Opaque,
    PolyFuncType,
    RowVariable,
    StringArg,
    StringParam,
    Sum,
    Tuple,
    TupleArg,
    TupleConcatArg,
    TupleParam,
    Type,
    TypeArg,
    TypeBound,
    TypeParam,
    TypeTypeArg,
    TypeTypeParam,
    Variable,
    _QubitDef,
)

ImportContext = model.Term | model.Node | model.Region | str


class ModelImportError(Exception):
    """Exception raised when importing from the model representation fails."""

    def __init__(self, message: str, location: ImportContext | None = None):
        self.message = message
        self.location = location

        match location:
            case model.Term() as term:
                location_error = f"Error caused by term:\n```\n{term}\n```"
            case model.Region() as region:
                location_error = f"Error caused by region:\n```\n{region}\n```"
            case model.Node() as node:
                location_error = f"Error caused by node:\n```\n{node}\n```"
            case str() as other:
                location_error = other
            case None:
                location_error = "Error in unspecified location."

        super().__init__(f"{message}\n{location_error}")


def _collect_meta_json(node: model.Node) -> NodeMetadata:
    """Collects the `core.meta_json` metadata on the given node."""
    metadata = {}

    for meta in node.meta:
        match meta:
            case model.Apply(
                "compat.meta_json",
                [model.Literal(str() as key), model.Literal(str() as value)],
            ):
                pass
            case _:
                continue

        try:
            decoded = json.loads(value)
        except json.JSONDecodeError as err:
            error = "Failed to decode JSON metadata."
            raise ModelImportError(error, node) from err

        metadata[key] = decoded

    return NodeMetadata(metadata)


def _find_meta_title(node: model.Node) -> str | None:
    """Searches for `core.title` metadata on the given node."""
    for meta in node.meta:
        match meta:
            case model.Apply("core.title", [model.Literal(str() as title)]):
                return title
            case model.Apply("core.title"):
                error = "Invalid instance of `core.title` metadata."
                raise ModelImportError(error, meta)
            case _:
                pass

    return None


def _find_meta_order_region(region: model.Region) -> "RegionOrderHints":
    """Searches for order hint metadata on the given region."""
    data = RegionOrderHints()

    for meta in region.meta:
        match meta:
            case model.Apply(
                "core.order_hint.input_key", [model.Literal(int() as key)]
            ):
                data.input_keys.append(key)
            case model.Apply(
                "core.order_hint.output_key", [model.Literal(int() as key)]
            ):
                data.output_keys.append(key)
            case model.Apply(
                "core.order_hint.order",
                [model.Literal(int() as before), model.Literal(int() as after)],
            ):
                data.edges.append((before, after))
            case _:
                pass

    return data


def _collect_meta_order_keys(node: model.Node) -> list[int]:
    """Collects all order hint keys in the metadata of a node."""
    keys = []

    for meta in node.meta:
        match meta:
            case model.Apply("core.order_hint.key", [model.Literal(int() as key)]):
                keys.append(key)
            case _:
                pass

    return keys


class ModelImport:
    """Helper to import a Hugr."""

    local_vars: dict[str, "LocalVarData"]
    current_symbol: str | None
    link_prefix: int | None
    linked_ports: dict[str, tuple[list[InPort], list[OutPort]]]
    static_edges: list[tuple[Node, Node]]

    module: model.Module
    symbols: dict[str, model.Node]
    fn_nodes: dict[str, Node]
    fn_calls: list[tuple[str, Node]]
    hugr: Hugr

    def __init__(self, module: model.Module):
        self.local_vars = {}
        self.current_symbol = None
        self.module = module
        self.symbols = {}
        self.hugr = Hugr()
        self.link_prefix = None
        self.linked_ports = {}
        self.static_edges = []
        self.fn_nodes = {}
        self.fn_calls = []

        for node in module.root.children:
            symbol_name = node.operation.symbol_name()

            if symbol_name is None:
                continue

            if symbol_name in self.symbols:
                error = f"Duplicate symbol name `{symbol_name}`."
                raise ModelImportError(error, node)

            self.symbols[symbol_name] = node

    def add_node(
        self, node: model.Node, operation: Op, parent: Node, num_outs: int | None = None
    ) -> Node:
        """Add a model Node to the Hugr and record its in- and out-links."""
        node_id = self.hugr.add_node(
            op=operation,
            parent=parent,
            num_outs=num_outs,
            metadata=_collect_meta_json(node),
        )
        self.record_in_links(node_id, node.inputs)
        self.record_out_links(node_id, node.outputs)
        if model.Apply("core.entrypoint") in node.meta:
            self.hugr.entrypoint = node_id
        return node_id

    def record_in_links(self, node: Node, links: Iterable[str]):
        """Record a bunch of links entering the given Hugr Node with the given names."""
        for offset, link in enumerate(links):
            in_port = InPort(node=node, offset=offset)
            self.linked_ports.setdefault(f"{self.link_prefix}_{link}", ([], []))[
                0
            ].append(in_port)

    def record_out_links(self, node: Node, links: Iterable[str]):
        """Record a bunch of links exiting the given Hugr Node with the given names."""
        for offset, link in enumerate(links):
            out_port = OutPort(node=node, offset=offset)
            self.linked_ports.setdefault(f"{self.link_prefix}_{link}", ([], []))[
                1
            ].append(out_port)

    def link_ports(self):
        """Add links to the Hugr according to the recorded data."""
        for link, (in_ports, out_ports) in self.linked_ports.items():
            match in_ports, out_ports:
                case [[], []]:
                    raise AssertionError
                case _, [out_port]:
                    for in_port in in_ports:
                        self.hugr.add_link(out_port, in_port)
                case [[in_port], _]:
                    for out_port in out_ports:
                        self.hugr.add_link(out_port, in_port)
                case _, _:
                    error = f"Link `{link}` has multiple inputs and outputs."
                    raise ModelImportError(error)

    def link_static_ports(self):
        for symbol, callnode in self.fn_calls:
            self.static_edges.append((self.fn_nodes[symbol], callnode))
        for src, dst in self.static_edges:
            out_port_offset = self.hugr.num_out_ports(src) - 1
            out_port = OutPort(node=src, offset=out_port_offset)

            in_port_offset = self.hugr.num_in_ports(dst)
            in_port = InPort(node=dst, offset=in_port_offset)

            self.hugr.add_link(out_port, in_port)

    def add_module_metadata(self):
        self.hugr[self.hugr.module_root].metadata = _collect_meta_json(self.module.root)

    def import_dfg_region(self, region: model.Region, parent: Node):
        """Import an entire DFG region from the model into the Hugr."""
        signature = self.import_signature(region.signature)

        input_node = self.hugr.add_node(
            Input(signature.input), parent=parent, num_outs=len(signature.input)
        )
        self.record_out_links(input_node, region.sources)

        output_node = self.hugr.add_node(Output(signature.output), parent=parent)
        self.record_in_links(output_node, region.targets)

        order_data = _find_meta_order_region(region)
        order_data.add_node_keys(input_node, order_data.input_keys)
        order_data.add_node_keys(output_node, order_data.output_keys)

        for child in region.children:
            child_id = self.import_node_in_dfg(child, parent)
            child_order_keys = _collect_meta_order_keys(child)
            order_data.add_node_keys(child_id, child_order_keys)

        for src_key, tgt_key in order_data.edges:
            src_node = order_data.get_node_by_key(src_key)
            tgt_node = order_data.get_node_by_key(tgt_key)
            self.hugr.add_order_link(src_node, tgt_node)

    def import_block(self, block: model.Node, parent: Node):
        # 1. Add the DataFlowBlock node:
        match block.signature:
            case model.Apply("core.ctrl", [ctrl_inputs, ctrl_outputs]):
                pass
            case _:
                error = f"Invalid signature for {block}."
                raise ModelImportError(error)
        match list(ctrl_inputs.to_list_parts()):
            case [inputs]:
                pass
            case _:
                error = f"DFB inputs should be singleton list: {ctrl_inputs}."
                raise ModelImportError(error)
        assert isinstance(inputs, model.Term)
        block_node = self.add_node(
            block,
            # TODO The translation here seems to be underdetermined. It could be
            # DataflowBlock(
            #     self.import_type_row(inputs),
            #     Sum(ts),
            #     ss,
            # ),
            # where the ctrl_outputs have been expressed as:
            # [[*ts[0], *ss], [*ts[1], *ss], ...]
            # with ss some common suffix of the lists in ctrl_outputs. But how do we
            # decide on that common suffix? Below we take it to be empty.
            DataflowBlock(
                self.import_type_row(inputs),
                Sum(
                    [
                        self.import_type_row(cast(model.Term, output))
                        for output in ctrl_outputs.to_list_parts()
                    ]
                ),
                [],
            ),
            parent,
        )
        # 2. Import the dataflow region:
        [block_region] = block.regions
        self.import_dfg_region(block_region, block_node)

    def import_cfg_region(
        self, region: model.Region, signature: FunctionType, parent: Node
    ):
        """Import an entire CFG region from the model into the Hugr."""
        [entry_link] = region.sources
        entry_block_idx = None
        for i, child in enumerate(region.children):
            if entry_link in child.inputs:
                entry_block_idx = i
                break
        assert entry_block_idx is not None
        entry_block = region.children[entry_block_idx]

        # 1. Import the entry block:
        self.import_block(entry_block, parent)

        # 2. Create the exit node:
        exit_node = self.hugr.add_node(ExitBlock(signature.output), parent)
        self.record_in_links(exit_node, region.targets)

        # 3. Import the other blocks:
        for i, child in enumerate(region.children):
            if i != entry_block_idx:
                self.import_block(child, parent)

    def import_node_in_dfg(self, node: model.Node, parent: Node) -> Node:
        """Import a model Node within a DFG region.

        Returns the Hugr Node corresponding to the model Node. The correspondence is
        almost 1-1, but a LoadConst model Node requires two Hugr Nodes (Const and
        LoadConst); in this case the LoadConst is returned.
        """
        signature = self.import_signature(node.signature)

        def import_dfg_node() -> Node:
            match node.regions:
                case [body]:
                    pass
                case _:
                    error = "DFG node expects a dataflow region."
                    raise ModelImportError(error, node)
            node_id = self.add_node(
                node, DFG(signature.input, signature.output), parent
            )
            self.import_dfg_region(body, node_id)
            return node_id

        def import_tail_loop() -> Node:
            match node.regions:
                case [body]:
                    pass
                case _:
                    error = "Loop node expects a dataflow region."
                    raise ModelImportError(error, node)

            match body.signature:
                case model.Apply("core.fn", [_, body_outputs]):
                    pass
                case _:
                    error = "Tail loop body expects `(core.fn _ _)` signature."
                    raise ModelImportError(error, node)

            match list(_import_closed_list(body_outputs)):
                case [model.Apply("core.adt", [variants]), *rest]:
                    pass
                case _:
                    error = "TailLoop body expects `(core.adt _)` as first target type."
                    raise ModelImportError(error, node)

            match list(_import_closed_list(variants)):
                case [just_inputs, just_outputs]:
                    pass
                case _:
                    error = "TailLoop body expects sum type with two variants."
                    raise ModelImportError(error, node)

            node_id = self.add_node(
                node,
                TailLoop(
                    just_inputs=self.import_type_row(just_inputs),
                    rest=[self.import_type(t) for t in rest],
                    _just_outputs=self.import_type_row(just_outputs),
                ),
                parent,
                len(signature.output),
            )
            self.import_dfg_region(body, node_id)
            return node_id

        def import_custom_node(op: model.Term) -> Node:
            match op:
                case model.Apply(symbol, args):
                    pass
                case _:
                    error = "The operation of a custom node must be a symbol "
                    "application."
                    raise ModelImportError(error, node)

            match symbol:
                case "core.call":
                    _input_types, _output_types, func = args
                    match func:
                        case model.Apply(fn_symbol, fn_args):
                            pass
                        case _:
                            error = "The function of a Call node must be a symbol "
                            "application."
                            raise ModelImportError(error, node)
                    type_args = [self.import_type_arg(fn_arg) for fn_arg in fn_args]
                    callnode = self.add_node(
                        node,
                        Call(
                            # FIXME PolyFuncType needs list[TypeParam], not
                            # list[TypeArg]. How to get this?
                            signature=PolyFuncType(type_args, signature),  # type: ignore[arg-type]
                            instantiation=signature,
                            type_args=type_args,
                        ),
                        parent,
                        len(signature.output),
                    )
                    self.fn_calls.append((fn_symbol, callnode))
                    return callnode
                case "core.call_indirect":
                    [inputs, outputs] = args
                    sig = FunctionType(
                        self.import_type_row(inputs), self.import_type_row(outputs)
                    )
                    callindirectnode = self.add_node(
                        node, CallIndirect(sig), parent, len(signature.output)
                    )
                    return callindirectnode
                case "core.load_const":
                    value = args[-1]
                    [datatype] = signature.output
                    match datatype:
                        case FunctionType(_inputs, _outputs):
                            # Import as a LoadFunc operation.
                            match value:
                                case model.Apply(str() as fn_id, fn_args):
                                    pass
                                case _:
                                    error = "Unexpected arguments to core.load_const: "
                                    f"{args}"
                                    raise ModelImportError(error, node)
                            type_args = [
                                self.import_type_arg(fn_arg) for fn_arg in fn_args
                            ]
                            loadfunc_node = self.add_node(
                                node,
                                LoadFunc(
                                    # FIXME PolyFuncType needs list[TypeParam], not
                                    # list[TypeArg]. How to get this?
                                    PolyFuncType(type_args, datatype),  # type: ignore[arg-type]
                                    datatype,
                                    type_args,
                                ),
                                parent,
                                1,
                            )
                            self.fn_calls.append((fn_id, loadfunc_node))
                            return loadfunc_node
                        case _:
                            # Import as a Const and a LoadConst node.
                            v = self.import_value(value)
                            const_node = self.hugr.add_node(Const(v), parent, 1)
                            loadconst_node = self.add_node(
                                node, LoadConst(datatype), parent, 1
                            )
                            self.hugr.add_link(
                                OutPort(const_node, 0), InPort(loadconst_node, 0)
                            )
                            return loadconst_node
                case "core.make_adt":
                    tag = args[-1]
                    match tag:
                        case model.Literal(int() as tagval):
                            pass
                        case _:
                            error = f"Unexpected tag: {tag}"
                            raise ModelImportError(error)
                    [sigout] = signature.output
                    match sigout:
                        case Sum(_variant_rows) as output_sum:
                            pass
                        case _:
                            error = f"Invalid signature with {symbol}: {node.signature}"
                            raise ModelImportError(error)
                    return self.add_node(node, Tag(tagval, output_sum), parent, 1)
                case "prelude.MakeTuple":
                    [arglist] = args
                    return self.add_node(
                        node,
                        MakeTuple(self.import_type_row(arglist)),
                        parent,
                        1,
                    )
                case "prelude.UnpackTuple":
                    [arglist] = args
                    typerow = self.import_type_row(arglist)
                    return self.add_node(
                        node,
                        UnpackTuple(typerow),
                        parent,
                        len(typerow),
                    )
                # Others are imported as Custom nodes.
                case _:
                    extension, op_name = _split_extension_name(symbol)
                    return self.add_node(
                        node,
                        Custom(
                            op_name=op_name,
                            extension=extension,
                            signature=signature,
                            args=[self.import_type_arg(arg) for arg in args],
                        ),
                        parent,
                        len(signature.output),
                    )

        def import_cfg() -> Node:
            match node.regions:
                case [body]:
                    pass
                case _:
                    error = "CFG node expects a control-flow region."
                    raise ModelImportError(error, node)
            node_id = self.add_node(
                node, CFG(signature.input, signature.output), parent
            )
            self.import_cfg_region(body, signature, node_id)
            return node_id

        def import_conditional() -> Node:
            match node.signature:
                case model.Apply("core.fn", [inputs, outputs]):
                    pass
                case _:
                    error = "Conditional node expects `(core.fn _ _)` signature."
                    raise ModelImportError(error, node)

            match list(_import_closed_list(inputs)):
                case [model.Apply("core.adt", [variants]), *other_inputs]:
                    sum_ty = Sum(
                        [
                            self.import_type_row(variant)
                            for variant in _import_closed_list(variants)
                        ]
                    )
                case _:
                    error = (
                        "Conditional node expects `(core.adt _)` as first input type."
                    )
                    raise ModelImportError(
                        error,
                        node,
                    )

            node_id = self.add_node(
                node,
                Conditional(
                    sum_ty=sum_ty,
                    other_inputs=[self.import_type(t) for t in other_inputs],
                    _outputs=self.import_type_row(outputs),
                ),
                parent,
            )

            for case_body in node.regions:
                case_signature = self.import_signature(case_body.signature)
                case_id = self.hugr.add_node(
                    Case(inputs=case_signature.input, _outputs=case_signature.output),
                    node_id,
                )
                self.import_dfg_region(case_body, case_id)

            return node_id

        match node.operation:
            case model.InvalidOp():
                error = "Invalid operation can not be imported."
                raise ModelImportError(error, node)
            case model.Dfg():
                return import_dfg_node()
            case model.Cfg():
                return import_cfg()
            case model.Block():
                error = "Unexpected basic block."
                raise ModelImportError(error, node)
            case model.CustomOp(op):
                return import_custom_node(op)
            case model.TailLoop():
                return import_tail_loop()
            case model.Conditional():
                return import_conditional()
            case _:
                error = "Unexpected node in DFG region."
                raise ModelImportError(error, node)

    def import_node_in_module(self, node: model.Node, link_prefix: int) -> Node | None:
        """Import a model Node at the Hugr Module level."""
        self.link_prefix = link_prefix

        def import_declare_func(symbol: model.Symbol) -> Node:
            f_name = _find_meta_title(node)
            if f_name is None:
                f_name = symbol.name
            signature = self.enter_symbol(symbol)
            node_id = self.add_node(
                node,
                FuncDecl(
                    f_name=f_name, signature=signature, visibility=symbol.visibility
                ),
                self.hugr.module_root,
                1,
            )
            self.exit_symbol()
            self.fn_nodes[symbol.name] = node_id
            return node_id

        def import_define_func(symbol: model.Symbol) -> Node:
            f_name = _find_meta_title(node)
            if f_name is None:
                f_name = symbol.name
            signature = self.enter_symbol(symbol)
            node_id = self.add_node(
                node,
                FuncDefn(
                    f_name=f_name,
                    inputs=signature.body.input,
                    _outputs=signature.body.output,
                    params=signature.params,
                    visibility=symbol.visibility,
                ),
                self.hugr.module_root,
                1,
            )

            match node.regions:
                case [body]:
                    pass
                case _:
                    error = "Function definition expects a single region."
                    raise ModelImportError(error, node)

            self.import_dfg_region(body, node_id)
            self.exit_symbol()
            self.fn_nodes[symbol.name] = node_id
            return node_id

        def import_declare_alias(symbol: model.Symbol) -> Node:
            match symbol:
                case model.Symbol(
                    name=name,
                    visibility=_visibility,
                    signature=model.Apply("core.type", []),
                ):
                    pass
                case _:
                    error = f"Unexpected symbol in alias declaration: {symbol}"
                    raise ModelImportError(error)
            return self.add_node(
                node,
                AliasDecl(alias=name, bound=TypeBound.Copyable),  # TODO which bound?
                self.hugr.module_root,
            )

        def import_define_alias(symbol: model.Symbol, value: model.Term) -> Node:
            match symbol:
                case model.Symbol(
                    name=name,
                    visibility=_visibility,
                    signature=model.Apply("core.type", []),
                ):
                    pass
                case _:
                    error = f"Unexpected symbol in alias definition: {symbol}"
                    raise ModelImportError(error)
            return self.add_node(
                node,
                AliasDefn(alias=name, definition=self.import_type(value)),
                self.hugr.module_root,
            )

        imported_node = None
        match node.operation:
            case model.DeclareFunc(symbol):
                imported_node = import_declare_func(symbol)
            case model.DefineFunc(symbol):
                imported_node = import_define_func(symbol)
            case model.DeclareAlias(symbol):
                imported_node = import_declare_alias(symbol)
            case model.DefineAlias(symbol, value):
                imported_node = import_define_alias(symbol, value)
            case model.Import():
                pass
            case model.DeclareConstructor():
                pass
            case model.DeclareOperation():
                pass
            case _:
                error = "Unexpected node in module region."
                raise ModelImportError(error, node)
        self.link_prefix = None
        return imported_node

    def enter_symbol(self, symbol: model.Symbol) -> PolyFuncType:
        assert len(self.local_vars) == 0

        bounds: dict[str, TypeBound] = {}

        for constraint in symbol.constraints:
            match constraint:
                case model.Apply("core.nonlinear", [model.Var(name)]):
                    bounds[name] = TypeBound.Copyable
                case _:
                    error = "Constraint other than `core.nonlinear` on a variable."
                    raise ModelImportError(error, constraint)

        param_types: list[TypeParam] = []

        for index, param in enumerate(symbol.params):
            bound = bounds.get(param.name, TypeBound.Linear)
            type = self.import_type_param(param.type, bound=bound)
            self.local_vars[param.name] = LocalVarData(index, type)
            param_types.append(type)

        body = self.import_signature(symbol.signature)
        return PolyFuncType(param_types, body)

    def exit_symbol(self):
        self.local_vars = {}

    def import_signature(self, term: model.Term | None) -> FunctionType:
        match term:
            case None:
                error = "Signature required."
                raise ModelImportError(error)
            case model.Apply("core.fn", [inputs, outputs]):
                return FunctionType(
                    self.import_type_row(inputs), self.import_type_row(outputs)
                )
            case _:
                error = "Invalid signature."
                raise ModelImportError(error, term)

    def lookup_var(self, name: str) -> "LocalVarData":
        if name not in self.local_vars:
            error = f"Unknown variable `{name}`."
            raise ModelImportError(error)

        return self.local_vars[name]

    def import_type_param(
        self, term: model.Term, bound: TypeBound = TypeBound.Linear
    ) -> TypeParam:
        """Import a TypeParam from a model Term."""
        match term:
            case model.Apply("core.nat"):
                return BoundedNatParam()
            case model.Apply("core.str"):
                return StringParam()
            case model.Apply("core.float"):
                return FloatParam()
            case model.Apply("core.bytes"):
                return BytesParam()
            case model.Apply("core.type"):
                return TypeTypeParam(bound)
            case model.Apply("core.list", [item_type]):
                return ListParam(self.import_type_param(item_type))
            case model.Apply("core.tuple", [item_types]):
                return TupleParam(
                    [
                        self.import_type_param(item_type)
                        for item_type in _import_closed_list(item_types)
                    ]
                )
            case model.Apply("core.const", [runtime_type]):
                return ConstParam(self.import_type(runtime_type))
            case _:
                error = "Failed to import TypeParam."
                raise ModelImportError(error, term)

    def import_type_arg(self, term: model.Term) -> TypeArg:
        """Import a TypeArg from a model Term."""

        def import_list(term: model.Term) -> TypeArg:
            lists: list[TypeArg] = []

            for group in _group_seq_parts(term.to_list_parts()):
                if isinstance(group, list):
                    lists.append(
                        ListArg([self.import_type_arg(item) for item in group])
                    )
                else:
                    lists.append(self.import_type_arg(group))

            return ListConcatArg(lists).flatten()

        def import_tuple(term: model.Term) -> TypeArg:
            tuples: list[TypeArg] = []

            for group in _group_seq_parts(term.to_list_parts()):
                if isinstance(group, list):
                    tuples.append(
                        TupleArg([self.import_type_arg(item) for item in group])
                    )
                else:
                    tuples.append(self.import_type_arg(group))

            return TupleConcatArg(tuples).flatten()

        match term:
            case model.Literal(str() as value):
                return StringArg(value)
            case model.Literal(int() as value):
                return BoundedNatArg(value)
            case model.Literal(float() as value):
                return FloatArg(value)
            case model.Literal(bytes() as value):
                return BytesArg(value)
            case model.List():
                return import_list(term)
            case model.Tuple():
                return import_tuple(term)
            case _:
                # Assume it's a TypeTypeArg
                return TypeTypeArg(self.import_type(term))

    def import_type(self, term: model.Term) -> Type:
        """Import the type from a model Term."""
        match term:
            case model.Apply("core.fn", [inputs, outputs]):
                return FunctionType(
                    self.import_type_row(inputs), self.import_type_row(outputs)
                )
            case model.Apply("core.adt", [variants]):
                return Sum(
                    [
                        self.import_type_row(variant)
                        for variant in _import_closed_list(variants)
                    ]
                )
            case model.Apply("prelude.qubit", []):
                return _QubitDef()
            case model.Apply(symbol, args):
                extension, type_id = _split_extension_name(symbol)
                return Opaque(
                    id=type_id,
                    extension=extension,
                    # TODO How to determine the type bound (Copyable or Linear)?
                    bound=TypeBound.Copyable,
                    args=[self.import_type_arg(arg) for arg in args],
                )
            case model.Var(name):
                var_data = self.lookup_var(name)
                return Variable(idx=var_data.index, bound=var_data.bound)
            case _:
                error = "Failed to import Type."
                raise ModelImportError(error, term)

    def import_type_row(self, term: model.Term) -> list[Type]:
        def import_part(part: model.SeqPart) -> Type:
            if isinstance(part, model.Splice):
                if isinstance(part.seq, model.Var):
                    var_data = self.lookup_var(part.seq.name)
                    return RowVariable(var_data.index, var_data.bound)
                else:
                    error = "Can only import spliced variables."
                    raise ModelImportError(error, term)
            else:
                return self.import_type(part)

        return [import_part(part) for part in term.to_list_parts()]

    def import_value(self, term: model.Term) -> val.Value:
        match term:
            case model.Apply(
                "arithmetic.int.const",
                [
                    model.Literal(int() as int_logwidth),
                    model.Literal(int() as int_value),
                ],
            ):
                # Ensure value is in signed form for conversion to IntVal:
                width = 1 << int_logwidth
                if int_value >= 1 << (width - 1):
                    int_value -= 1 << width
                return IntVal(int_value, int_logwidth)
            case model.Apply(
                "arithmetic.float.const_f64", [model.Literal(float() as float_value)]
            ):
                return FloatVal(float_value)
            case model.Apply("collections.array.const", [_, array_type, array_values]):
                return ArrayVal(
                    [
                        self.import_value(cast(model.Term, v))
                        for v in array_values.to_list_parts()
                    ],
                    self.import_type(array_type),
                )
            case model.Apply(
                "compat.const_json", [typ, model.Literal(str() as json_str)]
            ):
                json_dict = json.loads(json_str)
                match typ:
                    case model.Apply(typename, args):
                        match typename:
                            case "core.adt":
                                [arg] = args
                                match list(arg.to_list_parts()):
                                    case [model.List() as ts]:
                                        pass
                                    case _:
                                        error = f"Unexpected term: {term}"
                                        raise ModelImportError(error)
                                match json_dict:
                                    case {"c": "ConstExternalSymbol", "v": value}:
                                        return val.Extension(
                                            name="ConstExternalSymbol",
                                            typ=Tuple(
                                                *[
                                                    self.import_type(
                                                        cast(model.Term, t)
                                                    )
                                                    for t in ts.to_list_parts()
                                                ]
                                            ),
                                            val=value,
                                        )
                                    case _:
                                        error = f"Unexpected term: {term}"
                                        raise ModelImportError(error)
                            case _:
                                extension, type_id = _split_extension_name(typename)
                                match json_dict:
                                    case {"c": name, "v": value}:
                                        # Determine appropriate TypeBound
                                        bound = TypeBound.Copyable
                                        if typename == "collections.list.List":
                                            [arg] = args
                                            datatype = self.import_type(arg)
                                            bound = datatype.type_bound()
                                        # TODO Determine type bound in other cases
                                        return val.Extension(
                                            name=name,
                                            typ=Opaque(
                                                id=type_id,
                                                bound=bound,
                                                args=[
                                                    self.import_type_arg(arg)
                                                    for arg in args
                                                ],
                                                extension=extension,
                                            ),
                                            val=value,
                                        )
                                    case _:
                                        error = f"Unexpected term: {term}"
                                        raise ModelImportError(error)
                    case _:
                        error = f"Unexpected compat.const_json type: {typ}"
                        raise ModelImportError(error)
            case model.Apply("core.const.adt", [variants, _types, tag, values]):
                match tag:
                    case model.Literal(int() as tagval):
                        pass
                    case _:
                        error = f"Unexpected tag: {tag}"
                        raise ModelImportError(error)
                return val.Sum(
                    tag=tagval,
                    typ=Sum(
                        variant_rows=[
                            [
                                self.import_type(cast(model.Term, t))
                                for t in cast(model.Term, variant).to_list_parts()
                            ]
                            for variant in variants.to_list_parts()
                        ]
                    ),
                    vals=[
                        self.import_value(cast(model.Term, v))
                        for v in values.to_tuple_parts()
                    ],
                )
            case _:
                error = "Unsupported constant value."
                raise ModelImportError(error, term)


@dataclass
class LocalVarData:
    """Data describing a local variable."""

    index: int
    type: TypeParam
    bound: TypeBound = field(default=TypeBound.Linear)


@dataclass
class RegionOrderHints:
    """Order hint metadata."""

    input_keys: list[int] = field(default_factory=list)
    output_keys: list[int] = field(default_factory=list)
    edges: list[tuple[int, int]] = field(default_factory=list)
    key_to_node: dict[int, Node] = field(default_factory=dict)

    def add_node_keys(self, node: Node, keys: Iterable[int]):
        for key in keys:
            if key in self.key_to_node:
                error = f"Duplicate order key `{key}`."
                raise ModelImportError(error)

            self.key_to_node[key] = node

    def get_node_by_key(self, key: int) -> Node:
        if key not in self.key_to_node:
            error = f"Unknown order key `{key}`."
            raise ModelImportError(error)

        return self.key_to_node[key]


def _group_seq_parts(
    parts: Iterable[model.SeqPart],
) -> Generator[model.Term | list[model.Term]]:
    group: list[model.Term] = []

    for part in parts:
        if isinstance(part, model.Splice):
            if len(group) > 0:
                yield group
                group = []
            yield part.seq
        else:
            group.append(part)

    if len(group) > 0:
        yield group


def _import_closed_list(term: model.Term) -> Generator[model.Term]:
    for part in term.to_list_parts():
        if isinstance(part, model.Splice):
            error = "Expected closed list."
            raise ModelImportError(error, term)
        else:
            yield part


def _import_closed_tuple(term: model.Term) -> Generator[model.Term]:
    for part in term.to_tuple_parts():
        if isinstance(part, model.Splice):
            error = "Expected closed tuple."
            raise ModelImportError(error, term)
        else:
            yield part


def _split_extension_name(name: str) -> tuple[str, str]:
    match name.rsplit(".", 1):
        case [extension, id]:
            return (extension, id)
        case [id]:
            return ("", id)
        case _:
            raise AssertionError



"""HUGR package and pointed package interfaces."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Generic, TypeVar, cast

from typing_extensions import deprecated

import hugr._serialization.extension as ext_s
import hugr.model as model
from hugr import ext
from hugr.envelope import (
    EnvelopeConfig,
    _make_envelope,
    _make_envelope_str,
    read_envelope,
    read_envelope_str,
)
from hugr.hugr.base import Hugr
from hugr.ops import FuncDecl, FuncDefn, Op

if TYPE_CHECKING:
    from hugr.ext import Extension, ExtensionRegistry
    from hugr.hugr.node_port import Node

__all__ = [
    "Package",
    "PackagePointer",
    "ModulePointer",
    "ExtensionPointer",
    "NodePointer",
    "FuncDeclPointer",
    "FuncDefnPointer",
]


@dataclass(frozen=True)
class Package:
    """A package of HUGR modules and extensions.


    The HUGRs may refer to the included extensions or those not included.
    """

    #: HUGR modules in the package.
    modules: list[Hugr]
    #: Extensions included in the package.
    extensions: list[Extension] = field(default_factory=list)

    def _to_serial(self) -> ext_s.Package:
        return ext_s.Package(
            modules=[m._to_serial() for m in self.modules],
            extensions=[e._to_serial() for e in self.extensions],
        )

    @staticmethod
    def from_bytes(
        envelope: bytes, extensions: ExtensionRegistry | None = None
    ) -> Package:
        """Deserialize a byte string to a Package object.

        Some envelope formats can be read from a string. See :meth:`from_str`.

        Args:
            envelope: The byte string representing a Package.
            extensions: If not None, an extension registry to resolve the custom
                operations and types.

        Returns:
            The deserialized Package object.
        """
        package = read_envelope(envelope)
        if extensions is not None:
            package.used_extensions(extensions)
        return package

    @staticmethod
    def from_str(envelope: str, extensions: ExtensionRegistry | None = None) -> Package:
        """Deserialize a string to a Package object.

        Not all envelope formats can be read from a string.
        See :meth:`from_bytes` for a more general method.

        Args:
            envelope: The string representing a Package.
            extensions: If not None, an extension registry to resolve the custom
                operations and types.

        Returns:
            The deserialized Package object.
        """
        package = read_envelope_str(envelope)
        if extensions is not None:
            package.used_extensions(extensions)
        return package

    @staticmethod
    def from_model(package: model.Package, extensions: list[Extension] | None = None):
        if extensions is None:
            extensions = []
        return Package([Hugr.from_model(hugr) for hugr in package.modules], extensions)

    def to_bytes(self, config: EnvelopeConfig | None = None) -> bytes:
        """Serialize the package to a HUGR envelope byte string.

        Some envelope formats can be encoded into a string. See :meth:`to_str`.
        """
        config = config or EnvelopeConfig.BINARY
        return _make_envelope(self, config)

    def to_str(self, config: EnvelopeConfig | None = None) -> str:
        """Serialize the package to a HUGR envelope string.

        Not all envelope formats can be encoded into a string.
        See :meth:`to_bytes` for a more general method.
        """
        config = config or EnvelopeConfig.TEXT
        return _make_envelope_str(self, config)

    @deprecated("Use HUGR envelopes instead. See the `to_bytes` and `to_str` methods.")
    def to_json(self) -> str:
        """Serialize the package to a printable HUGR envelope string."""
        return self._to_serial().model_dump_json()

    @classmethod
    @deprecated(
        "Use HUGR envelopes instead. See the `from_bytes` and `from_str` methods."
    )
    def from_json(cls, json_str: str) -> Package:
        """Deserialize a JSON string to a Package object.

        Args:
            json_str: The JSON string representing a Package.

        Returns:
            The deserialized Package object.
        """
        return ext_s.Package.model_validate_json(json_str).deserialize()

    def to_model(self) -> model.Package:
        """Export the package as its hugr model representation.

        At the moment this does not yet contain the extensions.
        """
        return model.Package([module.to_model() for module in self.modules])

    def used_extensions(
        self, resolve_from: ext.ExtensionRegistry | None = None
    ) -> ext.ExtensionResolutionResult:
        """Get the extensions used by this Package, optionally resolving unresolved
        types and operations.

        This method modifies the HUGR modules and packaged extensions in-place when
        resolve_from is provided, replacing Custom operations with ExtOp operations and
        opaque types with ExtType when their extensions are found in the registry.

        Args:
            resolve_from: Optional extension registry to resolve against.
                If None, opaque types and Custom ops will not be resolved.

        Returns:
            The result of resolving the extensions, containing the used
            extensions and a list of referenced but unresolved extensions.

        Example:
            >>> from hugr.build import Dfg
            >>> Dfg(tys.Qubit).hugr.used_extensions().ids()
            {'prelude'}
        """
        from hugr.ext import ExtensionResolutionResult

        packaged_extensions = ext.ExtensionRegistry.from_extensions(self.extensions)

        if resolve_from is None:
            resolve_from = ext.ExtensionRegistry()
        resolve_from.extend(packaged_extensions)

        # Packaged extensions are always marked as "used".
        result = ExtensionResolutionResult(
            used_extensions=packaged_extensions,
        )

        for module in self.modules:
            result.extend(module.used_extensions(resolve_from))

        result._extend_with_transitive_ops(resolve_from)

        return result


@dataclass(frozen=True)
class PackagePointer:
    """Classes that point to packages and their inner contents."""

    #: Package pointed to.
    package: Package


@dataclass(frozen=True)
class ModulePointer(PackagePointer):
    """Pointer to a module in a package.

    Args:
        package: Package pointed to.
        module_index: Index of the module in the package.
    """

    #: Index of the module in the package.
    module_index: int

    @property
    def module(self) -> Hugr:
        """Hugr definition of the module."""
        return self.package.modules[self.module_index]

    def to_executable_package(self) -> ExecutablePackage:
        """Create an executable package from a module containing a main function.

        Raises:
            ValueError: If the module does not contain a main function.
        """
        module = self.module
        try:
            main_node = next(
                n
                for n in module.children()
                if isinstance((f_def := module[n].op), FuncDefn)
                and f_def.f_name == "main"
            )
        except StopIteration as e:
            msg = "Module does not contain a main function"
            raise ValueError(msg) from e
        return ExecutablePackage(self.package, self.module_index, main_node)


@dataclass(frozen=True)
class ExtensionPointer(PackagePointer):
    """Pointer to an extension in a package.

    Args:
        package: Package pointed to.
        extension_index: Index of the extension in the package.
    """

    #: Index of the extension in the package.
    extension_index: int

    @property
    def extension(self) -> Extension:
        """Extension definition."""
        return self.package.extensions[self.extension_index]


OpType = TypeVar("OpType", bound=Op)


@dataclass(frozen=True)
class NodePointer(Generic[OpType], ModulePointer):
    """Pointer to a node in a module.

    Args:
        package: Package pointed to.
        module_index: Index of the module in the package.
        node: Node pointed to
    """

    #: Node pointed to.
    node: Node

    @property
    def node_op(self) -> OpType:
        """Get the operation of the node."""
        return cast(OpType, self.module[self.node].op)


@dataclass(frozen=True)
class FuncDeclPointer(NodePointer[FuncDecl]):
    """Pointer to a function declaration in a module.

    Args:
        package: Package pointed to.
        module_index: Index of the module in the package.
        node: Node containing the function declaration.
    """

    @property
    def func_decl(self) -> FuncDecl:
        """Function declaration."""
        return self.node_op


@dataclass(frozen=True)
class FuncDefnPointer(NodePointer[FuncDefn]):
    """Pointer to a function definition in a module.

    Args:
        package: Package pointed to.
        module_index: Index of the module in the package.
        node: Node containing the function definition
    """

    @property
    def func_defn(self) -> FuncDefn:
        """Function definition."""
        return self.node_op


@dataclass(frozen=True)
class ExecutablePackage(FuncDefnPointer):
    """PackagePointer with a defined entrypoint node.

    Args:
        package: Package pointed to.
        module_index: Index of the module in the package.
        node: Node containing the entry point function definition.
    """

    @property
    def entry_point_node(self) -> Node:
        """Get the entry point node of the package."""
        return self.node
